#!/usr/bin/env python

import argparse
import csv
import importlib.resources
import mcvqoe
import os.path
import scipy.io.wavfile
import sys
import tempfile

#expected path components for csv files
csv_path_names = (('csv','2loc_tx-data'),('data'))

def make_parser():

    #-----------------------[Setup ArgumentParser object]-----------------------

    parser = argparse.ArgumentParser(
        description="Reprocess audio files and write a new .csv with newly measured values")
    parser.add_argument('datafile', default=None, type=str,
                        help='CSV file from test to reprocess')
    parser.add_argument('outfile', default=None, type=str, nargs='?',
                        help='file to write reprocessed CSV data to. Can be the same name as datafile to overwrite results. if omitted output will be written to stdout')
    parser.add_argument('-m', '--measurement', type=str, default=None, metavar='M',
                        help='measurement to use to do reprocessing')
    parser.add_argument('--audio-path', type=str, default=None, metavar='P', dest='audio_path',
                        help='Path to audio files for test. Will be found automatically if not given')
    parser.add_argument('-s', '--split-audio-folder', default='', type=str, dest='split_audio_dest',
                        help='Folder to store single word clips to')

    return parser

def get_module(module_name=None, datafile=None):

    if not module_name:

        #get module name
        module_name = mcvqoe.base.get_measurement_from_file(datafile)

        #make sure a module was found
        if not module_name:
            raise RuntimeError(f"Unable to determine measurement for '{datafile}'")
    else:
        #name given, clean up and use

        #make lowercase
        module_name = module_name.lower()

        #check if full import path was given
        if not module_name.startswith('mcvqoe.') :
            #add mcvqoe to the module include
            module_name = 'mcvqoe.' + module_name

    #load module and return
    return importlib.import_module(module_name).measure

def reprocess_file(test_obj, datafile, out_name, **kwargs):

    if not out_name:
        #split data file path into parts
        d,n=os.path.split(datafile)
        #construct new name for file
        out_name=os.path.join(d,'R'+n)

    print(f'Loading test data from \'{datafile}\'', file=sys.stderr)
    #read in test data
    test_dat=test_obj.load_test_data(datafile, **kwargs)

    print(f'Reprocessing test data to \'{out_name}\'', file=sys.stderr)

    test_obj.post_process(test_dat, out_name, test_obj.audio_path)

    return out_name


def main():

    #-----------------------------[Parse arguments]-----------------------------

    #get parser
    parser = make_parser()

    args = parser.parse_args()

    #---------------------------[Load Measure Class]---------------------------

    measurement_class = get_module(module_name=args.measurement, datafile=args.datafile)

    #---------------------------[Create Test object]---------------------------

    #create test obj to reprocess with
    test_obj=measurement_class()


    test_obj.split_audio_dest = args.split_audio_dest


    with tempfile.TemporaryDirectory() as tmp_dir:

        if(args.outfile=='--'):
            #print results, don't save file
            out_name=os.path.join(tmp_dir,'tmp.csv')
            print_outf=True
        else:
            out_name=args.outfile
            print_outf=False

        out_name = reprocess_file(test_obj, args.datafile, out_name, audio_path=args.audio_path)

        if(print_outf):
            with open(out_name,'rt') as out_file:
                dat=out_file.read()
            print(dat)
            print(f'Reprocessing complete', file=sys.stderr)
        else:
            print(f'Reprocessing complete for \'{out_name}\'', file=sys.stderr)


#main function
if __name__ == "__main__":
    main()


import warnings

#version info
from ..base.version import version

"""
Channel plugin for a clean channel with no impairments.

The clean channel plugin is the only one included with mcvqoe. It serves both as
a reference channel and as an example channel plugin for writing custom channel
plugins.

Attributes
----------
standard_delay: float
    The standard_delay attribute gives the delay that is added by the channel.
    This is used by QoEsim to compensate for the delay added by the channel and
    simulate the specified delay.
rates : list
    List of unique rates that can be used on the channel. This list should
    contain a list of possible rates. If there are multiple values that are
    accepted for a given rate, the one that is the most human readable should be
    chosen.
default_rate : str or number
    Default rate to use for channel. This should be one of the rates from
    `rates`.
"""

# plugin for clean channel

# No channel, no delay
standard_delay = 0

# no channel, no rate
default_rate = None
rates = []

channel_type = 'audio'

def simulate_audio_channel(
    tx_data, sample_rate, channel_rate=None, print_args=False, channel_impairment=None
):
    if channel_impairment:
        #apply impairment
        tx_data = channel_impairment(tx_data, sample_rate)
    if channel_rate:
        warnings.warn("For 'clean' there is no rate. 'channel_rate' option ignored")

    return tx_data



#!/usr/bin/env python

import random

import numpy as np

import pdb


def expected_psud(p_a, p_r, interval, message_length, method="EWC",
                  threshold=0.5):
    """
    Evaluate expected probability of successful delivery (PSuD).

    Evaluates the every word critical (EWC) PSuD, which requires an
    uninterrupted chain of keywords above the intelligibility threshold. Here
    the threshold is fixed at 0.5.

    Parameters
    ----------
    p_a : float
        Probability transitioning into a transmitting state.
    p_r : float
        Probability of remaining in a transmitting state.
    interval : float
        How often state transitions are evaluated.
    message_length : float
        Length of message in seconds.

    Returns
    -------
    float:
        Probability that the message was successfully delivered.

    Examples
    --------
    Evaluate the PSuD when P_a = P_r = 0.5, with an interval of 1 for a three
    second long message.

    >>> mcvqoe.simulation.expected_psud(0.5,0.5,1,3)
    """
    if(method == "EWC"):
        # TODO: Use threshold here...if it makes sense?
        psud = p_a * p_r ** (message_length/interval - 1)
    elif(method == "AMI"):
        # This can likely become a parameter later
        intell_threshold = threshold
        # Number of markov trials
        N = int(message_length/interval)

        # TODO: Add a check that if N > threshold use steady state
        # distriubtions rather than fancy brute forced probabilities

        psud = 0
        for k in range(2**N):
            # For each possible state of length N
            # String representation as 1s and 0s
            state = f'{k:0{N}b}'

            # Initialize state probability to 1
            state_prob = 1
            state_sum = 0
            for ix, x in enumerate(state):
                if(ix == 0 and x == '1'):
                    state_prob *= p_a
                    state_sum += 1
                elif(ix == 0 and x == '0'):
                    state_prob *= (1-p_a)
                elif(x == '1'):

                    if(state[ix-1] == '1'):
                        state_prob *= p_r
                    else:
                        state_prob *= p_a

                    # If success multiply state_probability by prob of success
                    # for markov trial ix
                    # state_prob *= p[ix]
                    state_sum += 1

                else:
                    if(state[ix-1] == '0'):
                        state_prob *= (1-p_a)
                    else:
                        state_prob *= (1-p_r)
                    # Otherwise multiply state_probability by prob of failure
                    # for markov trial ix
                    # state_prob *= (1-p[ix])

            if(state_sum/N >= intell_threshold):
                # Update psud with probability of success
                psud += state_prob

    else:
        raise ValueError(f"Unknown method passed: {method}.")
    return psud


class PBI:
    """
    Probabilityiser for simulating probablistic channel drops.

    This class allows simple simulations of a channel that has two states:
    transmitting and not transmitting. Using the probability of returning to
    the transmitting state given it is not transmitting, P_a, and the
    probability of remaining in the transmitting state given it is
    transmitting, P_r, a faulty channel can be simulated.

    Parameters
    ----------
    P_a1 : float, optional
        Probability of initializing the transmission. The default is 1.
    P_a2 : float, optional
        Probability of returning ot a transmitting state once the transmitting
        state has been reached and it has returned to a non-transmitting
        state. The default is None. In this case P_a2 is set to be equal to
        P_a1.
    P_r : float, optional
        Probability of remaining in a transmitting state. The default is 1.
    interval : float, optional
        How often the channel state is reevaluated, in seconds. The default is
        1.

    Attributes
    ----------
    P_a1 : float, optional
        Probability of initializing the transmission. The default is 1.
    P_a2 : float, optional
        Probability of returning ot a transmitting state once the transmitting
        state has been reached and it has returned to a non-transmitting
        state. The default is None. In this case P_a2 is set to be equal to
        P_a1.
    P_r : float, optional
        Probability of remaining in a transmitting state. The default is 1.
    interval : float, optional
        How often the channel state is reevaluated, in seconds. The default is
        1.
    state : int
        Current state of PBI, either INVALID, G0, G1, or H (0,1,2,3
        respectively).
    state_history : list
        List of transmitting states that the PBI has been in at each interval
        cycle. Values of 0 represent non-transmitting states, values of 1
        represent transmitting states.
    STATE_INVALID : int
        Value 0 to represent invalid state reached
    STATE_G0 : int
        Value 1 to represent in initial state where transmission has not
        started.
    STATE_G1 : int
        Value 2 to represent in non-transmitting state where transmission has
        occured at least once.
    STATE_H : int
        Value 3 to represent in transmitting state.

    Returns
    -------
    None.

    Examples
    --------
    >>> pb = mcvqoe.simulation.PBI(P_a1 = 0.5, P_r = 0.5)
    """

    STATE_INVALID = 0
    STATE_G0 = 1
    STATE_G1 = 2
    STATE_H = 3

    def __init__(self, P_a1=1, P_a2=None, P_r=1, interval=1):
        # time in seconds between state machine evaluations
        self.interval = interval
        self.initial_state()
        self.P_a1 = P_a1
        if P_a2 is None:
            self.P_a2 = self.P_a1
        else:
            self.P_a2 = P_a2
        self.P_r = P_r
        self.state_history = []
        
    def __repr__(self):
        return f'{type(self).__name__}({self.P_a1}, {self.P_a2}, {self.P_r}, {self.interval})'

    def initial_state(self):
        self.state = self.STATE_G0

    def process_audio(self, data, fs):
        # Numpy arrays pass by reference, and the result of
        # scipy.io.wavfile.read is read-only
        data = data.copy()

        # set to initial state
        self.initial_state()

        # calculate the number of samples in each chunk
        chunk_len = int(round(fs * self.interval))

        start = range(0, len(data), chunk_len)
        stop = list(range(chunk_len, len(data), chunk_len))
        # add end of array to stop
        stop.append(len(data))

        for s, e in zip(start, stop):
            self.update_state()
            if self.state != self.STATE_H:
                data[s:e] = 0
                self.state_history.append(0)
            else:
                self.state_history.append(1)

        return data

    def update_state(self):
        # generate random number for state transition
        r = random.random()
        # select next state based on current state
        if self.state == self.STATE_G0:
            if r < self.P_a1:
                # transition into H state
                self.state = self.STATE_H
            else:
                # stay in G0 state
                pass
        elif self.state == self.STATE_G1:
            if r < self.P_a2:
                # transition into H state
                self.state = self.STATE_H
            else:
                # stay in G1 state
                pass
        elif self.state == self.STATE_H:
            if r < self.P_r:
                # stay in H state
                pass
            else:
                # transition into G1 state
                self.state = self.STATE_G1

    def expected_psud(self, t):
        """
        Determine expected PSuD of message of length t given settings

        Convenience function to use expected_psud function given current PBI
        settings.

        Parameters
        ----------
        t : float
            Length of message in seconds.

        Returns
        -------
        psud : float
              PSuD of message of length t given current settings

        See Also
        --------
        expected_psud : function used by this convenience function.

        """

        psud = expected_psud(p_a=self.P_a1,
                             p_r=self.P_r,
                             interval=self.interval,
                             message_length=t)
        return psud



import scipy.signal
import sounddevice as sd
from fractions import Fraction
import mcvqoe.base
import tempfile
import pkgutil
import io
import os
import time


def single_play(ri, ap, audio_file=None, playback=False, ptt_wait=0.68, save_name=None):
    """
    Play an audio clip through a PTT system once.

    This function plays audio using the given RadioInterface and AudioPlayer
    objects. The output can, optionally, be played back to the usser when
    complete

    Parameters
    ----------
    ri : RadioInterface
        Interface to talk to radio PTT.
    ap : AudioPlayer
        Audio player to play audio with.
    audio_file : str, default=None
        Audio file to use for test. if None, use default audio.
    playback : bool, default=False
        If true, play the audio after the trial is complete.
    ptt_wait : float, default=0.68
        Amount of time to wait, in seconds, between PTT and playback.
    """

    if audio_file is None:
        audio_file = io.BytesIO(pkgutil.get_data("mcvqoe.audio_clips", "test.wav"))

    # get fs from audio player
    fs = ap.sample_rate

    # Gather audio data in numpy array and audio samplerate
    fs_file, audio_dat = mcvqoe.base.audio_read(audio_file)
    # Calculate resample factors
    rs_factor = Fraction(fs / fs_file)
    # Resample audio
    audio = scipy.signal.resample_poly(
        audio_dat, rs_factor.numerator, rs_factor.denominator
    )

    with tempfile.TemporaryDirectory() as tmp_dir:

        if save_name:
            # save in user specified location
            rec_file = save_name
        else:
            # set filename of recording
            rec_file = os.path.join(tmp_dir, "test.wav")

        # request access to channel
        ri.ptt(True)
        # pause for access
        time.sleep(ptt_wait)
        # play audio
        ap.play_record(audio, rec_file)
        # release channel
        ri.ptt(False)

        if playback:
            # read in file
            fs_rec, rec_dat = mcvqoe.base.audio_read(rec_file)
            # check if we have multiple channels
            if len(rec_dat.shape) > 1:
                # drop all but the first channel
                # this will drop the PTT tone if used and not blow eardrums
                rec_dat = rec_dat[:, 0]
            # set sample rate
            sd.default.samplerate = fs_rec
            # play audio over the default device
            sd.play(rec_dat)


# main function
def main():
    # need a couple of extra things for this to work
    # import them here
    import argparse
    import mcvqoe.hardware

    # -----------------------[Setup ArgumentParser object]-----------------------

    parser = argparse.ArgumentParser(description="Simple program to key radios")
    parser.add_argument(
        "-t",
        "--time",
        type=float,
        default=10,
        metavar="T",
        help="Time to key radio for in seconds",
    )
    parser.add_argument(
        "-r",
        "--radioport",
        default="",
        metavar="PORT",
        help="Port to use for radio interface. Defaults to the "
        + "first port where a radio interface is detected",
    )
    parser.add_argument(
        "-a", "--audio-file", default="", metavar="F", help="audio file to play"
    )
    parser.add_argument(
        "-p",
        "--overPlay",
        type=float,
        default=1,
        metavar="DUR",
        help="The number of seconds to play silence after the "
        + "audio is complete. This allows for all of the audio "
        + "to be recorded when there is delay in the system",
    )
    parser.add_argument(
        "-w",
        "--PTT-wait",
        type=float,
        default=0.68,
        metavar="T",
        dest="ptt_wait",
        help="Time to wait between pushing PTT and playing audio",
    )
    parser.add_argument(
        "-P",
        "--playback",
        default=False,
        action="store_true",
        help="Playback audio after it is recorded",
    )
    parser.add_argument(
        "-o", "--output", default=None, type=str,
        help="If given, the recorded audio will be saved into the filename provided." +
        "This filename must have a .wav extension and must not point to an existing file."
    )

    # -----------------------------[Parse arguments]-----------------------------

    args = parser.parse_args()

    # -----------------------------[set audio file]-----------------------------

    if args.audio_file:
        # load the file that was given
        f = args.audio_file
    else:
        # no audio given, use default
        f = None

    # ---------------------------[Create AudioPlayer]---------------------------

    ap = mcvqoe.hardware.AudioPlayer()

    # ---------------------------[Open RadioInterface]---------------------------

    with mcvqoe.hardware.RadioInterface(args.radioport) as ri:

        print(f"RadioInterface opened successfully on {ri.port_name}!")

        # --------------------------------[Key radio]--------------------------------

        print(f"Playing audio")

        single_play(
            ri,
            ap,
            audio_file=f,
            ptt_wait=args.ptt_wait,
            playback=args.playback,
            save_name=args.output,
        )

    print("Complete")


if __name__ == "__main__":
    main()




from warnings import warn

def terminal_progress_update(
            prog_type,
            num_trials,
            current_trial,
            msg="",
            clip_name="",
            delay="",
            file="",
            new_file="",
        ):
    if (prog_type == 'proc'):
        if (current_trial == 0):
            #we are post processing
            print('Processing test data')
        if (current_trial % 10 == 0):
            print(f'Processing trial {current_trial} of {num_trials}')   
    elif (prog_type == 'test'):
        if (current_trial == 0):
            print(f'Starting Test of {num_trials} trials')
        if (current_trial % 10 == 0):
            print(f'-----Trial {current_trial} of {num_trials}')
    elif (prog_type == 'compress'):
        if (current_trial == 0):
            print(f'Compressing {num_trials} audio files')
        if (current_trial % 10 == 0):
            print(f'Compressing trial {current_trial} of {num_trials}')
    elif (prog_type == 'warning'):
        warn(msg, stacklevel=2)
    elif (prog_type == 'check-fail'):
        print(f'On trial {current_trial} of {num_trials} : {msg}')
    elif (prog_type == 'check-resume'):
        print(f'Resuming test : {msg}')
    elif (prog_type == 'acc-clip-update'):
        print(f'---Delay : {delay:.3f}s\n'+
              f'---Clip  : {clip_name}')
    elif (prog_type == 'csv-update'):
        # Print name and location of datafile    
        print(f'--Starting {clip_name}\n'+
              f'--Storing data in: \'{file}\'')
    elif (prog_type == 'csv-rename'):
        print(f"Renaming '{file}' to '{new_file}'")
    elif prog_type == "status":
        print(msg)
        
    #continue test
    return True
    
def terminal_user_check(reason,message, trials=None, time=None):
    abort_test = False
    #check if we have time and trials
    if (trials and time):
        # Print set time
        print(f"Time for {trials} trials : {time}")
    #ask the user to press enter to continue
    #ring bell ('\a') to get their attention
    resp = input('\a'+
               message+' Press enter to continue\n'+
               'you may type "exit" to stop the test\n')
    if (resp.lower().strip() == 'exit'):
        abort_test = True
    
    return abort_test
    


import math
import mcvqoe.delay
import warnings

import numpy as np

from scipy import signal
from scipy.fft import fft

def fsf(tx_data, rx_data, fs=48E3):
    """
    Purpose
    -------
    Frequency slope fit distortion detector using Uniform Bands.
    
    Parameters
    ----------
    tx_data : numpy array
        Transmit audio data.
    rx_data : numpy array
        Received audio data.
    fs : double, optional
        Sample rate. The default is 48E3.
  
    Returns
    -------
    score : double
        FSF score.
    dly_samples : tuple or int
        M2E latency.

    """
    
    _, delay_samples = mcvqoe.delay.ITS_delay_est(rx_data, tx_data, mode='f', fs=48000)
    dly_samples = delay_samples
    if delay_samples < 0:
        #raise ValueError('Negative delay detected.')
        warnings.warn('Negative delay detected. Setting delay to zero.')
        delay_samples = 0
        
    if len(rx_data) < (len(tx_data) + delay_samples):
        rx_data = rx_data[delay_samples:]
    else:
        rx_data = rx_data[delay_samples:(len(tx_data) + delay_samples)]
    
    tx_slope = calc_slope(tx_data)
    
    rx_slope = calc_slope(rx_data)
        
    score = np.true_divide(rx_slope, tx_slope)
    
    return score, dly_samples
    
def calc_slope(wav_data):
    """
    Purpose
    -------
    Calculates frequency slope fit for a 
    given audio clip.
    
    Parameters
    ----------
    wav_data : numpy array
        Audio data.
    band_x : bool
        Determines whether or not linear fit x values
        come from band centers (false) or band numbers
        (true).

    Returns
    -------
    slope : double
        Slope of wav_data.

    """
    
    freq_set = np.array([[200, 450], 
                [400, 650], 
                [600, 850], 
                [800, 1050], 
                [1000, 1250],
                [1200, 1450],
                [1400, 1650], 
                [1600, 1850],
                [1800, 2050],
                [2000, 2250],
                [2200, 2450],
                [2400, 2650],
                [2600, 2850],
                [2800, 3050], 
                [3000, 3250]])
    
    fft_len = 2 ** 14
    
    num_bands = len(freq_set)

    #TODO: Check if already a column vector (avoid transpose if so)
    if wav_data.shape != (len(wav_data), ):
        wav_data = wav_data.T
    
    #TODO: Explore length of hamming window - see if length of window should be less than length of input wav data 
    # It seems that using the window in "signal.periodogram" is better than getting it here
    win = signal.get_window('hamming', len(wav_data))
  
    wav_win = win * wav_data
    
    # Perform periodogram function (equivalent to GNU Octave's built in periodogram)
    pxx, freq = periodogram(x=wav_data, window=win, nfft=fft_len, fs=int(48e3))
    
    band_vals = np.zeros((num_bands))
    
    for band in range(num_bands):
        mask = np.logical_and((freq >= freq_set[band, 0]), (freq <= freq_set[band, 1]))
        band_vals[band] = 10 * math.log10(np.mean(pxx[mask]))
  
    # Get index of max value, only consider the first half of FI bands
    max_idx = np.argmax(band_vals[0:(num_bands//2)])
    
    # Range of fit to the right of max
    fit_range = np.arange(max_idx, band_vals.shape[0])

    # x-axis values from band number
    x_vals = np.arange(0, num_bands)
    # Eliminate unused bands
    x_vals = x_vals[fit_range]

    # Using np.p.p.polyfit as is suggested in documentation
    p = np.polynomial.polynomial.polyfit(x_vals, band_vals[fit_range], 1)

    intercept = p[0]
    slope = p[1]

    return slope

def periodogram(x, window, nfft, fs):
    """
    Purpose
    -------
    Periodogram implementation that mimics Octave's implementation (periodogram_simple).

    Parameters
    ----------
    x : numpy array 
        Input signal.
    window : numpy array
        Window. It must be the same length as the input signal.
    nfft : int
        Number of DFT points.
    fs : int
        Sample rate. It must be positive.

    Returns
    -------
    None.

    """

    n = len(x)
    x = np.multiply(x, window)
    
    if n > nfft:
        rr = n % nfft
        
        if rr != 0:
            x = np.concatenate((x, np.zeros(nfft-rr)), axis=None)

        # Reshape uses 'order="F"' to match MATLAB version of reshape
        x = np.sum(np.reshape(x, (nfft, -1), order="F"), axis=1)
        
    n = np.sum(np.square(window))

    pxx = np.true_divide(np.square(np.abs(fft(x, n=nfft, axis=0))), n)
    pxx = np.true_divide(pxx, fs)
    
    # nfft is even
    if nfft % 2 == 0:
        psd_len = (nfft // 2) + 1
        tmp_arr = pxx[nfft-1: psd_len-1: -1]
        tmp_arr = np.append(tmp_arr, 0)
        tmp_arr = np.insert(tmp_arr, 0, 0)
        pxx = np.add(pxx[: psd_len], tmp_arr)

    # nfft is odd
    else:
        psd_len = (nfft + 1) // 2
        tmp_arr = pxx[nfft-1: psd_len-1: -1]
        tmp_arr = np.insert(tmp_arr, 0, 0)
        pxx = np.add(pxx[: psd_len], tmp_arr)
    
    freq = np.true_divide(np.arange(0, (nfft // 2)+1), nfft)
    freq = freq * fs

    return pxx, freq


import re

def audio_channels_to_string(channels):
    '''
    Convert a channels list to string for use in csv's.
    
    This is the revers of parse_audio_channels.
    
    Parameters
    ----------
    channels : list or tuple of strings
        List of strings describing the recorded channels as returned by
        `play_record`.
    
    Returns
    -------
    string
        a string representation of the channels
    
    See Also
    --------
    mcvqoe.hardware.audio_player : Hardware implementation of play_record.
    mcvqoe.hardware.QoEsim : Simulation implementation of play_record.
    mcvqoe.base.parse_audio_channels : Inverse of this function.
    '''
    
    if isinstance(channels,str):
        raise TypeError('channels can not be a string')
    
    #channel string
    return '('+(';'.join(channels))+')'


def parse_audio_channels(csv_str):
    '''
    Parse audio channels from string.
    
    This is the revers of audio_channels_to_string.
    
    Parameters
    ----------
    csv_str : string
        String with channel info as returned by audio_channels_to_string.
    
    Returns
    -------
    tuple of strings
        A tuple of strings similar to what would have been returned by
        `play_record`.
    
    See Also
    --------
    mcvqoe.hardware.audio_player : Hardware implementation of play_record.
    mcvqoe.hardware.QoEsim : Simulation implementation of play_record.
    mcvqoe.base.audio_channels_to_string : Inverse of this function.
    '''
    match=re.search('\((?P<channels>[^)]+)\)',csv_str)

    if(not match):
        raise ValueError(f'Unable to parse channels {csv_str}, expected in the form "(chan1;chan2;...)"')

    return tuple(match.group('channels').split(';'))


import io
import math
import re
from warnings import warn

import serial
import serial.tools.list_ports


# custom exception for commands
class CommandError(Exception):
    pass


class RadioInterface:
    """
    Class to talk to a radio interface board.
    
    Attributes
    ----------
    debug : bool, default=False
        If true, print extra info useful for debugging.
    default_radio : int, default=1

    See Also
    --------
    mcvqoe.simulation.QoEsim : Pretend to key fake radios.

    Examples
    --------
    Note that for these examples to work, you will need to have a radio interface
    plugged in.

    Create a radio interface object. This will search for a device on all serial
    ports on the machine.
    >>> ri = RadioInterface()

    Turn on an LED.
    >>> ri.led(1,'on')

    Key attached radio.
    >>> ri.ptt(True)

    Un-key attached radio.
    >>> ri.ptt(False)

    Turn off LED.
    >>> ri.led(1, 'off')

    If the serial port of the device is known it can be passed to the constructor.
    >>> ri = RadioInterface(port='COM12')
    """
    #USB PID/VID to use in search. Devices matching this PID/VID will be checked
    #This is a generic TI PID/VID that could get used by other things
    #TODO : do we want our own?
    _USB_VID = 0x2047 # RadioInterface Vendor ID (VID)
    _USB_PID = 0x0300 # RadioInterface Product ID (PID)

    def __init__(self, port=None, **kwargs):
        """
        Create a new RadioInterface object.

        to set class attributes, they can be specified with kwargs.

        Parameters
        ----------
        port : str, default=None
            Serial port to open to talk to radio interface hardware. If None is
            passed, all serial ports are searched for a device that seems to be
            connected to radio interface hardware. If a serial port string is
            given, it is used without sending any commands to it.
        """

        #set default values
        self.debug = False
        self.default_radio = 1

        #get properties from kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                #None value means keep defaults
                if v is not None:
                    setattr(self, k, v)
            else:
                raise TypeError(f"{k} is not a valid keyword argument")

        if not port:
            ports = serial.tools.list_ports.comports()
            for p in ports:
                #check for matching PID/VID
                if(not (p.vid==self._USB_VID and p.pid==self._USB_PID)):
                    if self.debug:
                        print(f"Skipping {p.device}, PID/VID does not match")
                    #skip this device
                    continue
                try:
                    self._openPort(p.device)
                    # send device type command
                    dt = self.devtype()

                    if self.debug:
                        print(f"{p.device} Devtype : {dt}")

                    if dt.startswith("MCV radio interface"):
                        # port found, done
                        break
                    else:
                        # close serial port
                        self.sobj.close()
                except (CommandError, serial.SerialException, UnicodeDecodeError) as e:
                    if self.debug:
                        print(e)
            else:
                # port not found, give error
                raise RuntimeError("No radio interface found")
        else:
            # open port
            self._openPort(port)

    def __repr__(self):
        string_props=('default_radio',)

        if hasattr(self, "sobj"):
            # check if port is open
            if self.sobj:
                #get the currently open port
                props=[f'port = {repr(self.port_name)}']

                for prop in string_props:
                    props.append(f'{prop} = {repr(getattr(self, prop))}')

                return f'{type(self).__name__}({", ".join(props)})'

        #otherwise port is not open give some sort of an indication
        return f'<inactive {type(self).__name__}>'


    def __enter__(self):

        return self

    def ptt(self, state, num=None):
        """
        Change the push-to-talk status of the radio interface.

        This will send a command to the radio interface to immediately change the
        status of the chosen PTT output.

        Parameters
        ----------
        state : bool
            If state is true then the radio will be keyed (transmitting). If
            state is False then the radio will not be keyed (not transmitting).
        num : int, default=None
            PTT output number to use. If None, self.default_radio is used.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.ptt_wait : Key radios after a delay.

        Examples
        --------

        Key the default radio.
        >>> ri.ptt(True)

        Un-key radio number 1.
        >>> ri.ptt(False,num=1)
        """

        if num is None:
            num = self.default_radio

        # check what the state is
        if state:
            # state is on
            state = "on"
        else:
            # state is off
            state = "off"

        # send command
        self._command(f"ptt {num} {state}")

    def led(self, num, state):
        """turn on or off LED's on the radio interface board

        LED(num,state) changes the state of the LED given by num. If state is
        true turn the LED on if state is False turn the LED off"""

        # determine LED state string
        if state:
            ststr = "on"
            if self.debug:
                print("state is on", flush=True)
        else:
            ststr = "off"
            if self.debug:
                print("state is off", flush=True)
        # send command
        self._command(f"LED {num} {ststr}")

    def devtype(self):
        """get the devicetype string from the radio interface

        dt = DEVTYPE() where dt is the devicetype string"""

        # flush input from buffer

        # send devtype command
        self._command("devtype")
        # get devtype line
        return self.textin.readline()

    def get_id(self):
        """Get the ID string from radio interface"""

        # Send ptt command with no arguments
        self._command("id")

        # Get response line
        ri_id = self.textin.readline()
        # Strip whitespace
        ri_id = ri_id.strip()

        return ri_id

    def get_version(self):
        """Get the version number from radio interface"""

        # Get devtype, this will have version
        devtype = self.devtype()
        parts = devtype.split(":")

        if len(parts) == 2:
            ver = parts[1]
            ver = ver.strip()
        else:
            ver = "old"
            warn("Unexpected devtype format, old RadioInterface firmware?")

        return ver

    def pttState(self):
        """returns the pttState for a radioInterface object. This is called
        automatically when pttState is accessed"""

        # flush input from buffer

        # send ptt command with no arguments
        self._command("ptt")
        # get response line
        resp = self.textin.readline()
        # extract number of PTT outputs
        m = re.match("(?P<num>\d+) PTT outputs:", resp)
        # check for match
        if not m:
            raise RuntimeError(f"could not parse '{resp}'")
        # parse outputs
        num = int(m.group("num"))
        # preallocate value
        value = num * [None]
        # loop through all outputs
        for k in range(num):
            # get response line
            resp = self.textin.readline()
            # get state from response
            m = re.match("PTT(\d+) status : (?:(?P<on>on)|(?P<off>off))", resp)
            # check that state was parsed correctly
            if m:
                if m.group("on"):
                    value[k] = True
                elif m.group("off"):
                    value[k] = False
                else:
                    value[k] = None
            else:
                value[k] = None
            #if there was a problem, print response in debug mode
            if value[k] is None and self.debug:
                print(f'Could not parse line : \'{resp}\'')

        return value

    def waitState(self):
        """returns the WaitState for a radioInterface object. this is called
        automatically when WaitState is accessed"""

        self._command("ptt state")
        # get response line
        resp = self.textin.readline()
        # parse PTT state
        m = re.match('PTT state : "(?P<state>.*?)"', resp)
        if m:
            return m.group("state")
        else:
            err_str = "Error : "
            if resp.startswith(err_str):
                raise RuntimeError(resp[len(err_str) :])
            else:
                raise RuntimeError(f"Unknown response '{resp}' received")

    def ptt_delay(self, delay, num=None, use_signal=False):
        """setup the radio interface to key the radio after a delay

        PTT_DELAY(dly) set the radio to be keyed in dly seconds.

        PTT_DELAY(dly,use_signal=True) set the radio to be keyed dly seconds
        after the start signal is detected.

        PTT_DELAY(dly,num=n,__) same as above but used key radio number n
        instead of the default radio

        delay=PTT_DELAY(dly,__) same as above but return the actual delay set on
        the microcontroller. This is different because of rounding and limits on
        the possible delay
        """

        if num is None:
            # No radio number given, use default
            num = self.default_radio

        # check which ptt command to use
        if use_signal:
            # delay is relative to signal
            delay_type = "Sdelay"
        else:
            # delay is relative to when command processed
            delay_type = "delay"

        # send ptt command with no arguments
        self._command(f"ptt {num} {delay_type} {delay}")
        # get response line
        resp = self.textin.readline()
        # get delay from string
        m = re.match(
            "(?:Error : (?P<err>.+)$)|(?:PTT in (?P<dly>[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?) sec)",
            resp,
        )

        # check for match
        if not m:
            raise RuntimeError(f"Unable to parse esponse {resp}")

        # check PTT delay was found
        if m.group("dly"):
            return float(m.group("dly"))
        elif m.group("err"):
            raise RuntimeError(m.group("err"))
        else:
            # unable to parse error
            raise RuntimeError("Unknown Error")

    def temp(self):
        """read value from temperature sensors

        [ext,int]=temp() - returns the temperature measured by the thermistor
        external to the radiointerface or the temperature sensor built into the
        MSP430
        """

        # send temp command
        self._command("temp")

        # get internal temp line
        intl = self.textin.readline()
        # get external temp line
        extl = self.textin.readline()

        # parse internal temperature
        m_int = re.match("int = (?P<temp>[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?) C", intl)
        # check for match
        if m_int:
            # get temperature value
            t_int = float(m_int.group("temp"))
        else:
            raise RuntimeError(f"could not parse internal temperature from '{intl}'")
        # parse external temp value
        m_ext = re.match("ext = (?P<val>\d+)", extl)
        # check for match
        if m_ext:
            # get integer ADC value
            v_ext = int(m_ext.group("val"))
        else:
            raise RuntimeError(f"could not parse external temperature from '{extl}'")
        # B value of thermistor
        B = 3470
        # compute external temperature
        t_ext = (
            B / math.log(10e3 / ((2 ** 12 - 1) / v_ext - 1) / (10e3 * math.exp(-B / (273.15 + 25))))
            - 273.15
        )

        return (t_int, t_ext)

    # delete method
    def __del__(self):
        
        #check if we have a serial object
        if hasattr(self, "sobj"):
            # check if port is open
            if self.sobj:
                # closeout command, turn off LEDS and ptt
                self.sobj.write(b"closeout\n")

                # wait for all data to write out
                self.sobj.flush()

            # flush and close text wrapper
            self.textin.close()

    def __exit__(self, exc_type, exc_value, exc_traceback):

        # check if we have a serial object
        if hasattr(self, "sobj"):
            # check if we had a serial problem
            if (not exc_type) or "serial" not in exc_type.__name__.lower():
                self._command("closeout")
            else:
                print("Problem with serial port detected, no cleanup performed")

    def _openPort(self, port):

        self.sobj = serial.Serial(port, timeout=0.5)
        # buffer serial data
        self.textin = io.TextIOWrapper(io.BufferedReader(self.sobj))

        #save the name of the open port so it can be used later
        self.port_name = port

    def _command(self, cmd):
        """low level command function to send command to the MSP430"""

        # trim extraneous white space from command
        cmd = cmd.strip()

        if self.debug:
            print(f"sending '{cmd}'")

        # assemble string and encode to bytes
        cmd_b = (cmd + "\n").encode("utf-8")
        # send command
        self.sobj.write(cmd_b)

        # line buffer
        l = ""

        # maximum number of iterations
        mi = 3

        # check command responses for echo
        while cmd not in l:
            # get response
            l = self.textin.readline()

            # strip whitespace
            l = l.strip()

            if self.debug and l:
                print(f"recived '{l}'")

            # subtract one from count
            mi -= 1
            # check if we should timeout
            if mi <= 0:
                # throw error
                raise CommandError("Command response timeout")



#!/usr/bin/env python

import argparse
import datetime
import configparser
import glob
import os
import re
import shutil

# prefix to show that this path needs sub folders copied
recur_prefix = "*"

# prefix to show not to backup sync folder
noback_prefix = "-"

def prog_str(prog_type, **kwargs):

    split_type = prog_type.split('-')

    major_type = split_type[0]

    if len(split_type) > 1:
        minor_type = split_type[1]
    else:
        #no minor type
        minor_type = None

    progress_str = ''

    if prog_type == 'main-section' :
        progress_str = f'Running section {kwargs["sect"]}'
    elif prog_type == 'log-complete':
        if kwargs['lines']:
            progress_str = f'{kwargs["lines"]} lines copied\n'
        else:
            progress_str = "Log files are identical, no lines copied\n"
        # add success message`
        progress_str += f'Log updated successfully to {kwargs["file"]}\n'
    #common things
    elif minor_type == 'dir':
        if major_type == 'log':
            progress_str = f'Finding Log files in \'{kwargs["dir"]}\''
        else:
            progress_str = f'Checking directory \'{kwargs["dir"]}\' for new files'
    elif minor_type == 'skip':
        if 'dir' in kwargs:
            progress_str = f'No new files found to copy to \'{kwargs["dir"]}\''
        else:
            progress_str = 'Up to date'
    elif minor_type == 'backup':
        progress_str = f'Backing files up from \'{kwargs["dest"]}\' to \'{kwargs["src"]}\''
    elif minor_type == 'invalid':
        progress_str = f'Skipping \'{kwargs["dir"]}\' it is not a directory or .zip file'
    elif minor_type == 'new':
        progress_str = f'Creating folder \'{kwargs["dir"]}\''
    elif minor_type == 'temp':
        progress_str = f'Skipping \'{kwargs["file"]}\''
    elif minor_type == 'skipdir':
        progress_str = f'Skipping Directory \'{kwargs["dir"]}\''
    elif minor_type == 'srcdest':
        progress_str = f'Copying \'{kwargs["src"]}\' to \'{kwargs["dest"]}\''
    #cull things
    elif prog_type == 'cull-deldir':
        progress_str = f'Deleting old directory \'{kwargs["dir"]}\''
    elif prog_type == 'cull-delfile':
        progress_str = f'Deleting old directory \'{kwargs["file"]}\''
    elif prog_type == 'cull-baddate' :
        progress_str = f'Unable to parse date in file \'{kwargs["file"]}\''
    elif prog_type == 'cull-badname' :
        progress_str = f'Unable to parse filename \'{kwargs["file"]}\''
    #skipping things
    elif prog_type == 'skip-later' :
        progress_str = f'Skipping {kwargs["file"]} for later'
    elif prog_type == 'skip-start' :
        #ignore indent here, this will be done at the end
        progress_str = f'Copying skipped {kwargs["ext"]} files'
    elif prog_type == 'supdate-old':
        progress_str = "Sync version old, updating"
    elif prog_type == 'supdate-vmissing':
        progress_str = "Sync version missing, updating"
    elif prog_type == 'supdate-missing':
        progress_str = "Sync directory not found, updating"
    elif prog_type == 'recur-found':
        progress_str ='\n'+f'Settings file found syncing \'{kwargs["dir"]}\':'
    elif prog_type == 'recur-error':
        progress_str = f'Error while syncing : {kwargs["err"]}'

    return progress_str, major_type, minor_type

def terminal_progress_update(
            prog_type,
            total,
            current,
            **kwargs):
    indents = {
        'main' : 0,
        'sub' : 1,
        'cull' : 1,
        'log' : 1,
        'subsub' : 2,
        'skip' : 1,
        'supdate' : 0,
        }

    p_str, major_type, minor_type  = prog_str(prog_type, **kwargs)

    if major_type in indents:
        num = indents[major_type]
        if major_type == prog_type:
            num -=1

        indent = '\t' * num
    else:
        indent = ''

    if p_str:
        print(indent+p_str)
    elif prog_type == 'main' :
        print(indent + f'processing directory {current} of {total}')
    elif minor_type == 'start':
        print(indent + f'Found {total} files to copy')
    elif minor_type == 'update':
        #only update terminal for subsub
        if major_type == 'subsub' :
            if current % 100 ==0:
                print(f'copying file {current} of {total}')
    elif prog_type == 'sub' :
        print(indent+f'processing subdirectory {current} of {total}')

# data directory names
data_dirs = (
    # 1 location data/processed data folder
    os.path.join("data", "csv"),
    recur_prefix + os.path.join("data", "wav"),
    # 2 location tx and rx raw data folders
    os.path.join("data", "2loc_rx-data"),
    recur_prefix + os.path.join("data", "2loc_tx-data"),
    # recovery and error
    os.path.join("data", "recovery"),
    os.path.join("data", "error"),
    # legacy .mat files
    noback_prefix + "data_matfiles",
)

# class to copy directories and skip some files
class cpyDelay:
    def __init__(self, ext, **kwargs):
        # initialize list of skipped files
        self.skip_list = []
        self.dly_ext = ext
        self.verbose = 0
        self.progress_update = terminal_progress_update

        #get properties from kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise TypeError(f"{k} is not a valid keyword argument")

    def ignore_func(self, src, dest):
        def _ignore(path, names):
            # get relative path into src directory
            relpath = os.path.relpath(path, src)

            # set of names to ignore
            ignore = set()
            # look at all names to see if they should be ignored
            for n, name in enumerate(names):
                # get extension
                (_, ext) = os.path.splitext(name)
                # check if extension should be ignored
                if ext == self.dly_ext:
                    # add file to list to ignore
                    ignore.add(name)
                    # construct full filenames to save for later
                    full_src = os.path.join(path, name)
                    full_dst = os.path.join(dest, relpath, name)
                    # add to list
                    self.skip_list.append((full_src, full_dst))
                    if self.verbose:
                        # give update
                        self.progress_update('skip-later', 0, 0, file=full_src)
            # return files found to ignore
            return ignore

        # return function to be used for ignore with copytree
        return _ignore

    def copytree(self, src, dest, **kwargs):
        # call copytree with ignore function
        shutil.copytree(src, dest, ignore=self.ignore_func(src, dest), **kwargs)

    def copy_skipped(self):
        # get number of files to copy
        total = len(self.skip_list)
        # notify user of delayed copies
        self.progress_update('skip-start', total, 0, ext=self.dly_ext)
        # loop through skipped files
        for n, (src, dest) in enumerate(self.skip_list):
            self.progress_update('skip-copy', total, n, src=src, dest=dest)
            shutil.copy2(src, dest)
        # clear out skip_list
        skip_list = []


class TestSyncer:
    def __init__(self, **kwargs):
        #set default values
        self.progress_update=terminal_progress_update
        self.bd=False
        self.cull=False
        self.sunset=30
        self.thorough=False
        self.copied_files = set()

        #get properties from kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise TypeError(f"{k} is not a valid keyword argument")

        # create object to skip .zip files with
        self.skip_zip = cpyDelay(".zip", verbose=True, progress_update=self.progress_update)

    def _dir_cpy(self, src, dest):
        # update progress
        self.progress_update('subsub-dir', 0, 0, dir=src)
        # set of files in source directory
        sset = set(os.listdir(src))
        # create dest if it does not exist
        os.makedirs(dest, exist_ok=True)
        # set of files in the destination directory
        dset = set(os.listdir(dest))
        # get the files that are not in dest
        cpy = sset.difference(dset)
        # find number of files to copy
        cnum = len(cpy)
        # check if we have files to copy
        if cnum:
            self.progress_update('subsub-start', cnum, 0)
            for count, f in enumerate(cpy, 1):
                # update progress
                self.progress_update('subsub-update', cnum, count)
                # create source and destination names
                sname = os.path.join(src, f)
                dname = os.path.join(dest, f)
                # copy file
                shutil.copyfile(sname, dname)
                # copy metadata
                shutil.copystat(sname, dname)

        else:
            self.progress_update('subsub-skip', 0, 0, dir=src)


    # function to copy unique files
    def sync_files(self, spath, dpath):
        # get datetime object for current time
        current_time = datetime.datetime.now()

        # loop over files in data_dirs
        for dat in data_dirs:
            # check for directory of directory path
            if dat.startswith(recur_prefix):
                recur = True
                # strip recur prefix from name
                dat_d = dat[len(recur_prefix) :]
            elif dat.startswith(noback_prefix):
                recur = False
                # strip recur prefix from name
                dat_d = dat[len(noback_prefix) :]
            else:
                recur = False
                dat_d = dat
            # construct source name
            src = os.path.join(spath, dat_d)
            # check if source dir exists
            if os.path.exists(src):
                self.progress_update('sub-dir', 0, 0, dir=src)
                # set of files in source directory
                sset = set(os.listdir(src))
                # construct destination name
                dest = os.path.join(dpath, dat_d)
                # check if directory name exists and source directory has files
                if sset and not os.path.exists(dest):
                    self.progress_update('sub-new', 0, 0, dir=dest)
                    # create folder
                    os.makedirs(dest, exist_ok=True)
                if os.path.exists(dest):
                    # set of files in the destination directory
                    dset = set(os.listdir(dest))
                else:
                    # empty set for dest
                    dset = set()
                # check if we are decending into directories
                if recur and self.thorough:
                    # yes, copy everything
                    # we will work out which files in subfolders to copy later

                    #get total number of dirs to check
                    cnum = len(sset)
                    for n, dir in enumerate(sset):
                        self.progress_update('sub-update', cnum, n, dir=sname)
                        # create source and destination names
                        sname = os.path.join(src, dir)
                        dname = os.path.join(dest, dir)
                        # check if this is a directory
                        if os.path.isdir(sname):
                            # yes, copy with dir_cpy
                            self._dir_cpy(sname, dname)
                            self.copied_files.add(os.path.basename(dname))
                        else:
                            # no, check extension
                            _, ext = os.path.splitext(sname)
                            # ignore case by lowering
                            ext = ext.lower()
                            # is this a zip file?
                            if ext == ".zip":
                                # yes, copy it
                                shutil.copyfile(sname, dname)
                                # and its metadata
                                shutil.copystat(sname, dname)
                                # add copied file name to set
                                simple_name, _ = os.path.splitext(os.path.basename(dname))
                                self.copied_files.add(simple_name)
                            else:
                                # no, give message
                                self.progress_update('sub-invalid', cnum, n, dir=sname)

                else:
                    # get the files that are not in dest
                    cpy = sset.difference(dset)
                    # check if there are files to copy
                    if cpy:
                        #get number of files to copy
                        cnum=len(cpy)
                        # copy files from src to dest that are not in dest
                        for n, f in enumerate(cpy):
                            #update progress
                            self.progress_update('sub-update', cnum, n, file=f)
                            # check for temp files
                            if f.endswith("TEMP.mat"):
                                self.progress_update('sub-temp', cnum, n, file=f)
                                # skip this file
                                continue
                            # create source and destination names
                            sname = os.path.join(src, f)
                            dname = os.path.join(dest, f)
                            # progress update with source and destination
                            self.progress_update('sub-srcdest', cnum, n, src=sname, dest=dname)
                            if recur and os.path.isdir(sname):
                                # copy directory
                                self.skip_zip.copytree(sname, dname)
                                # TODO: copy metadata?
                            else:
                                # copy file
                                shutil.copyfile(sname, dname)
                                # copy metadata
                                shutil.copystat(sname, dname)
                                # add copied file name to set
                                simple_name, _ = os.path.splitext(os.path.basename(dname))
                                self.copied_files.add(simple_name)
                    else:
                        self.progress_update('sub-skip', 0, 0, dir=dest)

                # check if we need to copy in the reverse
                if self.bd and not dat.startswith(noback_prefix):
                    # get the files that are not in src
                    cpy = dset.difference(sset)
                    # check if there are files to copy
                    if cpy:
                        # get number of files
                        cnum = len(cpy)
                        self.progress_update('sub-backup', cnum, 0, dest=dest, src=src)
                        # copy files from dest to src that are not in src
                        for n, f in enumerate(cpy):
                            self.progress_update('sub-update', cnum, n)
                            # check for temp files
                            if f.endswith("TEMP.mat"):
                                # print a message
                                self.progress_update('sub-temp', cnum, n, file=f)
                                # skip this file
                                continue
                            # create source and destination names
                            sname = os.path.join(dest, f)
                            dname = os.path.join(src, f)
                            # check if source is a directory
                            if not recur and os.path.isdir(sname):
                                # print a message
                                self.progress_update('sub-skipdir', cnum, n, dir=sname)
                                # skip this file
                                continue
                            # update progress
                            self.progress_update('sub-srcdest', cnum, n, src=sname, dest=dname)
                            if recur:
                                # copy directory
                                self.skip_zip.copytree(sname, dname)
                                # TODO: copy metadata?
                            else:
                                # copy file
                                shutil.copyfile(sname, dname)
                                # copy metadata
                                shutil.copystat(sname, dname)
                elif self.cull:
                    #get number of files
                    cnum=len(sset)
                    # find old files and delete them
                    for n, f in enumerate(sset):
                        self.progress_update('cull-update', cnum, n)

                        #big nasty regex to detect all the parts
                        fp_re = r'[A-z0-9]+_(?P<type>.*)' \
                                r'_(?P<date>\d{2}-[A-z][a-z]{2}-\d{2,4})' \
                                r'_(?P<time>\d{2}-\d{2}-\d{2})' \
                                r'(?:_(?P<word>[MF]\d_b\d+_w\d_[a-z]+))?' \
                                r'(?:_(?P<suffix>BAD|TEMP))?'
                        m = re.match(fp_re, f)

                        if not m:
                            self.progress_update('cull-badname', cnum, n, file=os.path.join(src, f))
                            #nothing more to do here
                            continue;

                        # grab date/time from filename
                        dstr = "_".join(m.group('date','time'))
                        try:
                            # parse string
                            f_date = datetime.datetime.strptime(dstr, "%d-%b-%Y_%H-%M-%S")
                            # calculate file age
                            age = current_time - f_date
                            # check if file is older than sunset days
                            if age.days > self.sunset:
                                # create file name
                                fullname = os.path.join(src, f)
                                # check for directory
                                if os.path.isdir(fullname):
                                    # print message
                                    self.progress_update('cull-dir', cnum, n, dir=fullname)
                                    shutil.rmtree(fullname)
                                else:
                                    # print message
                                    self.progress_update('cull-file', cnum, n, file=fullname)
                                    # delete file
                                    os.remove(fullname)
                        except ValueError:
                            self.progress_update('cull-baddate', cnum, n, file=os.path.join(src, f))

    def finish_tasks(self):
        # copy .zip files that we skipped earlier
        self.skip_zip.copy_skipped()

def export_sync(config_name, progress_update=terminal_progress_update, **kwargs):

    #don't allow cull to be used here
    if 'cull' in kwargs and kwargs['cull']:
        raise TypeError('cull is not supported for export_sync')

   # create a config parser
    config = configparser.ConfigParser()

    # load config file
    config.read(config_name)

    if not config.sections():
        raise ValueError(f"Configuration file not found at {config_name}.")

    # find configuration file location, all paths are relative to this
    config_fold = os.path.dirname(os.path.abspath(config_name))

    #create a sync object
    sync_obj = TestSyncer(progress_update=progress_update, **kwargs)

    #get number of sections
    snum = len(config.sections())

    for n, section in enumerate(config.sections()):
        # print section message
        progress_update('main-section', snum, n, sect=section)

        # get, what should be, the relative path from config
        src_rel = config[section]["src"]

        if os.path.isabs(src_rel):
            raise RuntimeError("source paths must not be absolute paths")

        # make path absolute
        src_dir = os.path.normpath(os.path.join(config_fold, src_rel))

        # print log files message
        progress_update('log-dir', 0, 0, dir=src_dir)
        # list log files
        logs = glob.glob(os.path.join(src_dir, "*.log"))

        # get number of logs
        lnum = len(logs)

        for ln, l in enumerate(logs):
            # get file name from path
            (h, name) = os.path.split(l)
            # generate destination filename
            dest = os.path.join(config[section]["dest"], name)
            progress_update('log-srcdest', lnum, ln, src=name, dest=dest)
            # make sure directory exists
            os.makedirs(os.path.dirname(dest), exist_ok=True)
            # copy log file into new directory
            shutil.copyfile(l, dest)
            # copy metadata
            shutil.copystat(l, dest)

        sync_obj.sync_files(src_dir, config[section]["dest"])

    # do things that are saved till the end
    sync_obj.finish_tasks()

    #return the set of copied files
    return sync_obj.copied_files

def import_sync(src, dest, progress_update=terminal_progress_update, **kwargs):
    #create a sync object
    sync_obj = TestSyncer(progress_update=progress_update, **kwargs)
    # print message
    progress_update('main-srcdest', 0, 0, src=src, dest=dest)
    # call sync function, don't copy data in the revers direction
    sync_obj.sync_files(src, dest)
    # do things that are saved till the end
    sync_obj.finish_tasks()

    #return the set of copied files
    return sync_obj.copied_files

# main function
def main():
    # get path name that this file is in
    file_path = os.path.dirname(os.path.abspath(__file__))

    parser = argparse.ArgumentParser(
        description="Copy MCV data files between drive and computers"
    )
    parser.add_argument(
        "- i",
        "--import",
        dest="imp",
        action="store",
        default=None,
        metavar="DIR",
        nargs=2,
        help="Use to import data from a given project folder",
    )
    parser.add_argument(
        "--cull",
        dest="cull",
        action="store_true",
        default=False,
        help="Remove old files from source directory",
    )
    parser.add_argument(
        "--no-cull",
        dest="cull",
        action="store_false",
        help="Do not remove old files from source directory",
    )
    parser.add_argument(
        "--sunset",
        dest="sunset",
        type=int,
        default=30,
        help="Delete files older than sunset days",
    )
    parser.add_argument(
        "--superficial",
        dest="thorough",
        action="store_false",
        default=False,
        help="Don't decend into datadir subfolders to check if new files exist",
    )
    parser.add_argument(
        "--thorough",
        dest="thorough",
        action="store_true",
        default=False,
        help="Decend into datadir subfolders to check if new files exist",
    )
    parser.add_argument(
        "-C",
        "--config",
        default=os.path.join(file_path, "testCpy.cfg"),
        metavar="CFG",
        dest="config",
        help="Path to configuration file (defaults to %(default)s)",
    )
    parser.add_argument(
        "--bidirectional",
        action='store_true',
        dest="bd",
        help="sync in both directions",
    )
    parser.add_argument(
        "--one-direction",
        action='store_false',
        dest="bd",
        help="sync only in one direction",
    )

    args = parser.parse_args()

    test_sync_arg_names = ('bd', 'thorough', 'cull', 'sunset')
    #extract arguments for TestSyncer
    test_sync_args = {k:v for k,v in vars(args).items() if k in test_sync_arg_names}

    # check if import argument was given
    if args.imp is None:
        files = export_sync(args.config, **test_sync_args)
        print(f'The following tests were copied : {" ,".join(files)}')
    else:
        # get source and destination folders from argument
        src = args.imp[0]
        dest = args.imp[1]
        # call import sync
        import_sync(src, dest, **test_sync_args)

if __name__ == "__main__":
    main()


import contextlib
import datetime
import math
import os
import queue
import sys
import threading
from fractions import Fraction

import numpy as np
import scipy.io.wavfile
import scipy.signal
import sounddevice as sd
import soundfile as sf

from mcvqoe.timing.soft_timecode import soft_time_fmt

try:
    import ctypes
    import threading

    class ThreadRecStop:
        """
        Class for stopping audio recording

        This class uses a thread running `input()` to get input from the user.
        This a fallback RecStop method that should be available on most platforms

        Parameters
        ----------

        Attributes
        ----------

        See Also
        --------
        mcvqoe.hardware.AudioPlayer : ThreadRecStop works with AudioPlayer.record

        Examples
        --------
        Record audio

        >>> import mcvqoe.hardware.AudioPlayer
        >>> ap=mcvqoe.hardware.AudioPlayer(fs=int(48e3))
        >>> ap.record('test.wav')

        """

        def _input(self):
            print("Press enter to stop")
            # wait for input
            input()
            # user pressed enter, done
            self._done = True

        def __enter__(self):
            self._done = False
            self.thread = threading.Thread(target=self._input, name="Console_Rec_input")
            self.thread.start()
            return self

        def is_done(self):
            """
            Method to check if recording should stop

            Returns
            -------
            bool
                True if recording should stop, False otherwise
            """
            return self._done

        def __exit__(self, exc_type, exc_value, exc_traceback):
            if self.thread.is_alive():
                thread_id = self.thread.get_ident()
                res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                    thread_id, ctypes.py_object(SystemExit)
                )
                if res > 1:
                    ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
                    print("Error while stopping thread")
            return False


except:
    ThreadRecStop = None

try:
    import msvcrt

    class WinRecStop:
        """
        Class for stopping audio recording

        This class uses msvcrt to get keypresses. This will only work on Windows.

        Parameters
        ----------

        Attributes
        ----------

        See Also
        --------
        mcvqoe.hardware.AudioPlayer : WinRecStop works with AudioPlayer.record

        Examples
        --------
        Record audio using WinRecStop to stop the recording

        >>> import mcvqoe.hardware.AudioPlayer
        >>> ap=mcvqoe.hardware.AudioPlayer(fs=int(48e3))
        >>> ap.record('test.wav',rec_stop=WinRecStop())

        """

        def __enter__(self):
            print("press any key to stop")
            return self

        def __exit__(self, exc_type, exc_value, exc_traceback):
            return False

        def is_done(self):
            """
            Method to check if recording should stop

            Returns
            -------
            bool
                True if recording should stop, False otherwise
            """
            if msvcrt.kbhit():
                # flush all keys
                while msvcrt.kbhit():
                    msvcrt.getch()
                # key pressed, we are done
                return True
            return False


except:
    # there was a problem, set to None
    WinRecStop = None

try:
    import sys
    import termios
    import tty

    class TermiosRecStop:
        """
        Class for stopping audio recording

        This class uses termios.tcsetattr to put the terminal into raw mode to
        be able to check for characters in the input without blocking.
        The termios library is only available on Unix systems.

        Parameters
        ----------

        Attributes
        ----------

        See Also
        --------
        mcvqoe.hardware.AudioPlayer : TermiosRecStop works with AudioPlayer.record

        Examples
        --------
        Record audio using TermiosRecStop to stop the recording

        >>> import mcvqoe.hardware.AudioPlayer
        >>> ap=mcvqoe.hardware.AudioPlayer(fs=int(48e3))
        >>> ap.record('test.wav',rec_stop=TermiosRecStop())

        """

        def __enter__(self):
            print("press any key to stop")
            # get stdin file descriptor
            self.fd = sys.stdin.fileno()
            # save old settings to restore later
            self.old_settings = termios.tcgetattr(self.fd)
            # change settings for raw input
            tty.setraw(sys.stdin.fileno())
            return self

        def __exit__(self, exc_type, exc_value, exc_traceback):
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
            return False

        def is_done(self):
            """
            Method to check if recording should stop

            Returns
            -------
            bool
                True if recording should stop, False otherwise
            """
            ch = sys.stdin.read(1)
            if len(ch) == 0:
                return True
            else:
                return False


except:
    TermiosRecStop = None

# set a sensible default for recording stop
if TermiosRecStop:
    DefaultRecStop = TermiosRecStop
elif WinRecStop:
    DefaultRecStop = WinRecStop
elif ThreadRecStop:
    DefaultRecStop = ThreadRecStop
else:
    # fallback to null context (terminate with ^C)
    DefaultRecStop = contextlib.nullcontext()


class AudioPlayer:
    """
    Class to play and record audio for test purposes.

    This class has functions for playing and recording audio and is used in QoE
    testing.

    Parameters
    ----------
    fs : int
        Sample rate of audio in/out in samples per second.
    blocksize : int
        The size of the blocks that are sent/received to/from the audio device.
    buffersize : int
        The number of blocks in the output buffer.
    overplay : float
        The number of seconds of extra audio to play/record at the end of a clip.
    rec_chans : dict
        Dictionary describing the recording. Dictionary keys should be one of
        {'rx_voice','PTT_signal','IRIGB_timecode','tx_beep','soft_timecode'}. The
        value for each entry is the, zero based, channel number that should be
        recorded for each signal. For the special 'soft_timecode' channel, no
        audio recording channel is used and the channel number should be 0.
    playback_chans : dict
        Dictionary describing the playback channels. Dictionary keys must be one
        of {'tx_voice','start_signal'}. The value for each entry is the, zero
        based, channel number that each signal should be played on.
    rec_stop : ContextManager
        Context manager used to test when a recording should be stopped.
        `rec_stop` is used as a context manager that is entered when the recording
        starts and exited when the recording stops. The `is_done` method is called
        to check if the recording should be terminated.

    Attributes
    ----------
    sample_rate : int
        Sample rate of audio in/out in samples per second.
    blocksize : int
        The size of the blocks that are sent/received to/from the audio device.
    buffersize : int
        The number of blocks in the output buffer.
    overplay : float
        The number of seconds of extra audio to play/record at the end of a clip.
    device : str
        Audio device to use. can be found with find_device().
    rec_chans : dict
        Dictionary describing the recording. Dictionary keys should be one of
        {'rx_voice','PTT_signal','IRIGB_timecode','tx_beep','soft_timecode'}. The
        value for each entry is the, zero based, channel number that should be
        recorded for each signal. For the special 'soft_timecode' channel, no
        audio recording channel is used and the channel number should be 0.
    playback_chans : dict
        Dictionary describing the playback channels. Dictionary keys must be one
        of {'tx_voice','start_signal'}. The value for each entry is the, zero
        based, channel number that each signal should be played on.
    rec_stop : ContextManager
        Context manager used to test when a recording should be stopped.
        `rec_stop` is used as a context manager that is entered when the recording
        starts and exited when the recording stops. The `is_done` method is called
        to check if the recording should be terminated.

    See Also
    --------
    mcvqoe.simulation.QoEsim : Simulated replacement for AudioPlayer.

    Examples
    --------

    play 48 kHz audio stored in tx_voice and record in a file named 'test.wav'.

    >>> import mcvqoe.hardware.AudioPlayer
    >>> ap=mcvqoe.hardware.AudioPlayer(fs=int(48e3))
    >>> ap.play_record(tx_voice,'test.wav')

    now do the same but also output the start signal on channel 1 and record the
    PTT signal on channel 1.

    >>> ap.playback_chans={'tx_voice':0,'start_signal':1}
    >>> ap.rec_chans={'rx_voice':0,'PTT_signal':1}
    >>> ap.play_record(tx_voice,'test.wav')
    """

    def __init__(
        self,
        device_str="UMC",
        **kwargs
    ):

        self.sample_rate = int(48e3)
        self.blocksize = 512
        self.buffersize = 20
        self.overplay = 1.0
        self.rec_chans = {"rx_voice": 0}
        self.playback_chans = {"tx_voice": 0}
        self.rec_stop = DefaultRecStop()


        #get properties from kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            #check if the 'fs' argument was given
            elif k == 'fs':
                #fs gets translated to sample_rate for legacy purposes
                self.sample_rate = v
            else:
                raise TypeError(f"{k} is not a valid keyword argument")

        try:
            self.device = self.find_device(device_str)
        except RuntimeError:
            #no device found
            self.device = None

    @staticmethod
    def find_device(device_str="UMC"):

        devs = sd.query_devices()

        for d in devs:
            if (
                d["max_input_channels"] > 0
                and d["max_output_channels"] > 0
                and device_str in d["name"]
            ):
                return d["name"]
        else:
            raise RuntimeError('No suitable audio interfaces found')

    def __repr__(self):
        string_props=('device','sample_rate','overplay','rec_chans','playback_chans')

        props=[]

        for prop in string_props:
            props.append(f'{prop} = {repr(getattr(self, prop))}')

        return f'{type(self).__name__}({", ".join(props)})'

    def record(self, filename):
        """
        Record audio based on rec_chans.

        Parameters
        ----------
        filename : str
            The file name to write audio to.

        """

        #check that device is set
        if not self.device:
            RuntimeError(f'No audio device found')

        # get the highest numbered channel
        # this will be the number of channels that will be played
        # account for zero based indexing
        chans = max(self.rec_chans.values()) + 1

        (rec_map, rec_names) = self._get_recording_map()

        # check if we are "recording" time data
        if "soft_timecode" in self.rec_chans:
            self._time_encode = True
            # find soft timecode channel
            soft_idx = rec_names.index("soft_timecode")
            # soft timecode will be appended
            # it's index will be the number of channels
            rec_map[soft_idx] = chans
        else:
            self._time_encode = False

        # Queue for recording input
        self._qr = queue.Queue()

        # set loop var to Fals (keep looping)
        rec_done = False

        # Make sure the file is opened before recording anything:
        with sf.SoundFile(
            filename, mode="x", samplerate=self.sample_rate, channels=len(rec_names)
        ) as file:
            with self.rec_stop, sd.InputStream(
                samplerate=self.sample_rate,
                device=self.device,
                channels=chans,
                callback=self._cb_rec,
            ):

                while not rec_done:
                    rx_dat = self._qr.get()
                    file.write(rx_dat[:, rec_map])

                    # check if we are done
                    rec_done = getattr(self.rec_stop, "is_done", lambda: False)()

            # Make sure to write any audio data still left in the recording queue
            while self._qr.empty() != True:
                rx_dat = self._qr.get()
                file.write(rx_dat[:, rec_map])

        # return the channels in the order recorded in the file
        return rec_names

    def _get_recording_map(self):
        """
        Get map and names for recording channels.

        Returns
        -------
        list of ints
            A channel map of the channel numbers used for each recording channel.
        list of strings
            A list of the names of the channels in the order they will be in in
            the recording file.
        """
        chan_map = []
        chan_names = []
        for k, v in self.rec_chans.items():
            chan_map.append(v)
            chan_names.append(k)
        return (chan_map, chan_names)

    def play_record(self, tx_voice, filename):
        """
        Play audio out the specified channels and record to 'filename'.

        Plays the audio specified by self.playback_chans on the respective
        channels. Records on the channels specified by self.rec_chans.

        Parameters
        ----------
        tx_voice : numpy array
            Voice audio to play through system. Needs to be in proper sample rate.
        filename : str
            The file extension to write audio data to.

        Returns
        -------
        list of strings
            A list of the recorded output channels in the order that they appear
            in the output file.

        See Also
        --------
        mcvqoe.simulation.QoEsim : Simulated replacement play_record.

        Examples
        --------

        play 48 kHz audio stored in tx_voice and record in a file named
        'test.wav'.

        >>> import mcvqoe.hardware.AudioPlayer
        >>> ap=mcvqoe.hardware.AudioPlayer(fs=int(48e3))
        >>> ap.play_record(tx_voice,'test.wav')

        now do the same but also output the start signal on channel 1 and record
        the PTT signal on channel 1.

        >>> ap.playback_chans={'tx_voice':0,'start_signal':1}
        >>> ap.rec_chans={'rx_voice':0,'PTT_signal':1}
        >>> ap.play_record(tx_voice,'test.wav')
        """

        #check that device is set
        if not self.device:
            RuntimeError(f'No audio device found')

        if len(tx_voice.shape) == 2:
            if tx_voice.shape[1] != 1:
                # TODO : warn about dropping channels
                pass
            # only take one channel of input
            tx_voice = tx_voice[:, 0]

        audio = np.empty((tx_voice.shape[0], len(self.playback_chans)))

        # get the highest numbered playback channel
        # this will be the number of channels that will be played
        # account for zero based indexing
        pb_chan = max(self.playback_chans.values()) + 1
        rec_chan = max(self.rec_chans.values()) + 1

        self._playback_map = []
        self._playback_silent = set(range(pb_chan))

        for n, (k, v) in enumerate(self.playback_chans.items()):
            # add to playback map
            self._playback_map.append(v)
            # remove from silent map
            self._playback_silent.remove(v)
            # Add start signal to audio
            if k == "start_signal":
                # Signal frequency
                f_sig = 1e3
                # Signal time
                t_sig = 22e-3
                # Calculate time for playback
                t = np.arange(float(len(audio))) / self.sample_rate
                # Calculate clip start signal
                sig = np.where((t < t_sig), np.sin(2 * np.pi * 1000 * t), 0)
                # Add start signal to audio
                audio[:, n] = sig
            elif k == "tx_voice":
                audio[:, n] = tx_voice
            else:
                raise ValueError(f"Unknown output channel : {k}")

        # convert to tuple for calback usage
        self._playback_silent = tuple(self._playback_silent)

        (rec_map, rec_names) = self._get_recording_map()

        # check if we are "recording" time data
        if "soft_timecode" in self.rec_chans:
            self._time_encode = True
            # find soft timecode channel
            soft_idx = rec_names.index("soft_timecode")
            # soft timecode will be appended
            # it's index will be the number of channels
            rec_map[soft_idx] = rec_chan
        else:
            self._time_encode = False

        # check for 1D audio
        if len(audio.shape) == 1:
            # promote a 1D array to a 2D nx1 array
            audio.shape = (audio.shape[0], 1)

        # Add Overplay
        if self.overplay != 0:
            audio = np.pad(
                audio,
                ((0, int(self.sample_rate * self.overplay)), (0, 0)),
                mode="constant",
            )

        # Queue for recording input
        self._qr = queue.Queue()
        # Queue for playback output
        self._qpb = queue.Queue(maxsize=self.buffersize)

        # Thread for callback function
        event = threading.Event()

        # NumPy audio array placeholder
        arr_place = 0

        for x in range(self.buffersize):

            data_slice = audio[
                self.blocksize * x : (self.blocksize * x) + self.blocksize
            ]

            if data_slice.size == 0:
                break

            # Save place of NumPy array slice for next loop
            arr_place += self.blocksize

            # Pre-fill queue
            self._qpb.put_nowait(data_slice)

        # Output and input stream in one
        # Latency of zero to try and cut down delay
        stream = sd.Stream(
            blocksize=self.blocksize,
            samplerate=self.sample_rate,
            dtype="float32",
            device=self.device,
            channels=[rec_chan, pb_chan],
            callback=self._cb_play_rec,
            finished_callback=event.set,
            latency=0,
        )

        with sf.SoundFile(
            filename,
            mode="x",
            samplerate=self.sample_rate,
            channels=len(self.rec_chans),
        ) as rec_file:
            with stream:
                timeout = self.blocksize * self.buffersize / self.sample_rate

                # For grabbing next blocksize slice of the NumPy audio array
                itrr = 0

                while data_slice.size != 0:

                    data_slice = audio[
                        arr_place
                        + (self.blocksize * itrr) : arr_place
                        + (self.blocksize * itrr)
                        + self.blocksize
                    ]
                    itrr += 1

                    self._qpb.put(data_slice, timeout=timeout)

                    rx_dat = self._qr.get()
                    rec_file.write(rx_dat[:, rec_map])
                # Wait until playback is finished
                event.wait()

            # Make sure to write any audio data still left in the recording queue
            while self._qr.empty() != True:
                rx_dat = self._qr.get()
                rec_file.write(rx_dat[:, rec_map])

        # return the channels in the order recorded in the file
        return rec_names

    def _cb_play_rec(self, indata, outdata, frames, time, status):
        """
        Callback function for the stream.
        Will run as long as there is audio data to play.

        """
        if self._time_encode:
            # get time
            tobj = datetime.datetime.utcnow()
            tobj -= datetime.timedelta(
                seconds=time.currentTime - time.inputBufferAdcTime
            )
            tstr = tobj.strftime(soft_time_fmt)
            dat_len = indata.shape[0]
            # make new column for array
            tcode = np.empty((dat_len, 1), dtype=indata.dtype)
            # check for floating point types
            if np.issubdtype(indata.dtype, np.floating):
                # scale 8 bit char values to full range
                scale = 1 / (255.0)
            else:
                scale = 1
            # check if timecode can fit in our chunk
            if dat_len >= len(tstr):
                for i in range(len(tstr)):
                    tcode[i] = ord(tstr[i]) * scale
                tcode[len(tstr) :] = 0
            else:
                tcode.fill(0xFF)
            # queue values
            self._qr.put_nowait(np.hstack((indata, tcode)))
        else:
            # Record the output
            self._qr.put_nowait(indata.copy())

        if status.output_underflow:
            print("Output underflow: increase blocksize?", file=sys.stderr)
            raise sd.CallbackStop
        assert not status
        try:
            data = self._qpb.get_nowait()
        except queue.Empty:
            print("Buffer is empty: increase buffersize?", file=sys.stderr)
            raise sd.CallbackStop

        if data.shape[0] < outdata.shape[0]:
            outdata[: data.shape[0], self._playback_map] = data
            outdata[: data.shape[0], self._playback_silent] = 0
            outdata[data.shape[0] :, :] = 0
            raise sd.CallbackStop
        else:
            outdata[:, self._playback_map] = data
            outdata[:, self._playback_silent] = 0

    def _cb_rec(self, indata, frames, time, status):
        """
        Callback function for the stream.
        Will run as long as there is audio data to play.

        """

        if status:
            print(status, file=sys.stderr, flush=True)

        if self._time_encode:
            # get time
            tobj = datetime.datetime.utcnow()
            tobj -= datetime.timedelta(
                seconds=time.currentTime - time.inputBufferAdcTime
            )
            tstr = tobj.strftime(soft_time_fmt)
            dat_len = indata.shape[0]
            # make new column for array
            tcode = np.empty((dat_len, 1), dtype=indata.dtype)
            # check for floating point types
            if np.issubdtype(indata.dtype, np.floating):
                # scale 8 bit char values to full range
                scale = 1 / (255.0)
            else:
                scale = 1
            # check if timecode can fit in our chunk
            if dat_len >= len(tstr):
                for i in range(len(tstr)):
                    tcode[i] = ord(tstr[i]) * scale
                tcode[len(tstr) :] = 0
            else:
                tcode.fill(0xFF)
            # queue values
            self._qr.put_nowait(np.hstack((indata, tcode)))
        else:
            self._qr.put(indata.copy())



from .version import version
from .misc import *
from .write_log import pre, post
from .audio_channels import *
from .fsf import *
from .naming import get_meas_basename, get_access_clip_info, get_meas_date_str, match_name
from .measurement import Measure



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 13 14:07:46 2021

@author: jkp4
"""
import appdirs
import argparse
import configparser
import os
import re
import shutil
import warnings

import pdb

appname = "mcvqoe"
appauthor = "MCV"


def local_copy(test_names, test_type, local_path=None,
               network_path=None, tx_wav=False, rx_wav = False):
    """
    Copy MCV QoE Measurement Data.

    Stores local_path and network_path for each test_type in a config file so
    they only need to be set once.

    Parameters
    ----------
    test_names : list
        Basenames of  tests.
    test_type : str
        Type of test (e.g. access, m2e, psud).
    local_path : str, optional
        Path where local data should be copied to. If None is passed,
        local_path is found from a config file. It must have been initialized
        for the given test type previously.
    network_path : str, optional
        Path where network data is copied from. If None is passed,
        network_path is found from a config file. It must have been initialzed
        for the given test type previously.
    tx_wav : bool
        Indicate whether or not to copy transmit audio wav file associated
        with cutpoints.

    Returns
    -------
    None.

    """
    # ---------------------[Load Config File]----------------------------------
    # Make application directories if they do not exist
    os.makedirs(appdirs.user_data_dir(appname, appauthor), exist_ok=True)

    # create a config parser
    config = configparser.ConfigParser()

    # load config file
    config_path = os.path.join(appdirs.user_data_dir(appname, appauthor),
                               "config.ini")
    config.read(config_path)

    if network_path is None:
        # No network path passed, pull from config file if we can
        if config.has_option(test_type, "network_path"):
            network_path = config[test_type]["network_path"]
        else:
            raise ValueError(
                "No network path passed and has not been configured"
                )
    else:
        # If new network path passed update config file
        if config.has_section(test_type):
            config[test_type]["network_path"] = network_path
        else:
            config[test_type] = {"network_path": network_path}

    if local_path is None:
        # No local path passed, pull from config file if we can
        if config.has_option(test_type, "local_path"):
            local_path = config[test_type]["local_path"]
        else:
            raise ValueError(
                "No local path passed and has not been configured"
                )
    else:
        # If new local path passed update config file
        if config.has_section(test_type):
            config[test_type]["local_path"] = local_path
        else:
            # Don't think we can hit this because network path goes first...
            config[test_type] = {"local_path": local_path}
    # ---------------------[Write config file]---------------------------------
    with open(config_path, "w") as configfile:
        config.write(configfile)

    # Make local path if it doesn't exist
    os.makedirs(local_path, exist_ok=True)

    # ---------------------[Ensure local path has proper subdirs]--------------
    if not os.path.exists(os.path.join(local_path, "csv")):
        os.makedirs(os.path.join(local_path, "csv"))
    if not os.path.exists(os.path.join(local_path, "wav")):
        os.makedirs(os.path.join(local_path, "wav"))
    # --------------------[Test CSV files]-------------------------------------
    csv_path = os.path.join(network_path, "csv")
    csv_files = os.listdir(csv_path)

    for cfile in csv_files:
        cpath = os.path.join(csv_path, cfile)
        for tname in test_names:
            if tname in cfile:
                lpath = os.path.join(local_path, "csv", cfile)
                if not os.path.exists(lpath):
                    print(f"Copying {cpath} to {lpath}")
                    shutil.copyfile(cpath, lpath)
                else:
                    print(f"Found locally: {lpath}")

    # --------------------[Cutpoint Files]----------------------------------
    for tname in test_names:
        # Remmove R if reprocessed data
        tname = tname.replace("Rcapture", "capture")
        # Path to wav folder
        wav_path = os.path.join(network_path, "wav", tname)
        if not os.path.exists(wav_path):
            warnings.warn(f"Test path not found in wav on network {wav_path}")
        else:
            # Get all files in wav_path
            wav_files = os.listdir(wav_path)

            # Define local path to store wav files, make it if need be
            lppath = os.path.join(local_path, "wav", tname)
            os.makedirs(lppath, exist_ok=True)

            # Search pattern to determine if looking at received file
            rx_pattern = r'^Rx\d+_.+.wav$'
            for wfile in wav_files:
                if ".csv" in wfile:
                    # Copy cut point file
                    wpath = os.path.join(wav_path, wfile)
                    lpath = os.path.join(lppath, wfile)
                    if not os.path.exists(lpath):
                        print(f"Copying {wpath} to {lpath}")
                        shutil.copyfile(wpath, lpath)
                    else:
                        print(f"Found locally: {lpath}")
                    # ------------[Transmit Audio]-----------------------------
                    # Copy Tx wav file if option set
                    if tx_wav:
                        wfile_no_ext, _ = os.path.splitext(wfile)
                        # Remove R is reprocessed data
                        tx_wav_wfile = wfile_no_ext + ".wav"

                        # Copy from path
                        tx_wpath = os.path.join(wav_path, tx_wav_wfile)
                        # Copy to path
                        tx_lpath = os.path.join(lppath, tx_wav_wfile)
                        # Check if file already exists
                        if not os.path.exists(tx_lpath):
                            print(f"Copying {tx_wpath} to {tx_lpath}")
                            shutil.copyfile(tx_wpath, tx_lpath)
                        else:
                            print(f"Found locally: {tx_lpath}")
                elif rx_wav and re.search(rx_pattern, wfile):
                    # Looking at a received audio file
                    # Copy from path
                    rx_wpath = os.path.join(wav_path, wfile)
                    # Copy to path
                    rx_lpath = os.path.join(lppath, wfile)
                    # Check if local file already exists
                    if not os.path.exists(rx_lpath):
                        print(f"Copying {rx_wpath} to {rx_lpath}")
                        shutil.copyfile(rx_wpath, rx_lpath)
                    else:
                        print(f"Found locally: {rx_lpath}")


def print_config(test_type=None):
    """
    Print the test_types and paths stored by local_copy.

    Parameters
    ----------
    test_type : str, optional
        If given, prints paths only for that test_type. The default is None.

    Returns
    -------
    None.

    """
    # ---------------------[Load Config File]----------------------------------
    # Make application directories if they do not exist
    os.makedirs(appdirs.user_data_dir(appname, appauthor), exist_ok=True)

    # create a config parser
    config = configparser.ConfigParser()

    # load config file
    config_path = os.path.join(appdirs.user_data_dir(appname, appauthor),
                               "config.ini")
    config.read(config_path)

    ttype_str = "----{}----"
    path_str = "{}: {}"
    if test_type is not None:
        print(ttype_str.format(test_type))
        for kv in config[test_type].keys():
            print(path_str.format(kv, config[test_type][kv]))
    else:
        for tt in config.keys():
            print(ttype_str.format(tt))
            for kv in config[tt].keys():
                print(path_str.format(kv, config[tt][kv]))


def convert_log_search_names(fnames):
    """
    Get log search ready file names.

    Convert output from mcvqoe.utilities.log_search datafilenames() to what
    local-copy expects.

    Parameters
    ----------
    fnames : TYPE
        DESCRIPTION.

    Returns
    -------
    lc_names : TYPE
        DESCRIPTION.

    """
    if type(fnames) is str:
        fnames = [fnames]
    lc_names = []
    for fn in fnames:
        _, short_name = os.path.split(fn)
        # TODO: Make this work with an optional non-capturing group for audio
        # file...
        # thought that '(capture_.+)(?:_\w{2}_b\d{1,2}_w\d{1}_\w+)?(?:.csv)'
        # should work but it doesn't
        # re_search = '(capture_.+)(?:_\w{2}_b\d{1,2}_w\d{1}_\w+)(?:.csv)'
        re_search = r"(R?capture_.+\d{2}-\d{2}-\d{2}).*.csv"
        # Extract part of the name that local-copy expects
        rs = re.search(re_search, short_name)
        if rs:
            f_id = rs.groups()
            lc_names.append(f_id[0])

    return lc_names


def main():
    """Command line method to run local-copy."""
    parser = argparse.ArgumentParser(
        description="Copy data files from network drive to local machine"
    )
    parser.add_argument("-f", "--test-names",
                        default=[],
                        type=str,
                        nargs="+",
                        help="Tests to copy"
                        )

    parser.add_argument("-t", "--test-type",
                        default=None,
                        type=str,
                        help="Type of test to copy from"
                        )
    parser.add_argument("-l", "--local-path",
                        default=None,
                        type=None,
                        help=("Local path to copy to. If not passed read from "
                              "last path used for this test-type."),
                        )

    parser.add_argument("-n", "--network-path",
                        default=None,
                        type=str,
                        help=("Network path to copy from. If not passed read "
                              "from last path used for this test-type."),
                        )
    parser.add_argument("-p", "--print-config",
                        default=False,
                        action="store_true",
                        help="Print current config file settings",
                        )

    parser.add_argument("-w", "--tx-wav",
                        default=False,
                        action="store_true",
                        help="Copy transmit audio wav file with cutpoints",
                        )
    parser.add_argument('-r', '--rx-wav',
                        default=False,
                        action="store_true",
                        help="Copy received audio wav file with cutpoints")

    args = parser.parse_args()

    if args.print_config:
        print_config(test_type=args.test_type)
    if args.test_names != []:

        if args.test_type is None and not args.print_config:
            raise ValueError("test_type is required")

        local_copy(
            args.test_names,
            test_type=args.test_type,
            local_path=args.local_path,
            network_path=args.network_path,
            tx_wav=args.tx_wav,
            rx_wav=args.rx_wav
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 08:20:53 2021

@author: cjg2
"""
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State

import json
import numpy as np
import os
import pandas as pd
import tempfile 

from mcvqoe.hub.eval_app import app

import mcvqoe.hub.eval_shared as eval_shared

#-----------------------[Begin layout]---------------------------

measurement = 'diagnostics'
layout = eval_shared.layout_template(measurement)

def format_diagnostic_results(m2e_eval, digits=6):
    """
    Format results from Test_Stats to be in HTML.

    Parameters
    ----------
    diagnostics_eval : 
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.
    """
    
   RX_Name = eval_shared.pretty_numbers(m2e_eval.mean, digits)
   A_Weight = eval_shared.pretty_numbers(m2e_eval.ci, digits)
   FSF_Scores
   children = html.Div([
        html.H6('Rx Clip Name'),
        html.Div(f'{pretty_mean} seconds'),
        html.H6('A Weight'),
        html.Div(f'{pretty_ci} dBA'),
        html.H6('Thinning factor'),
        html.Div(f'{m2e_eval.common_thinning}'),
        html.Div('TODO: Do something smarter when common thinning factor is not found'),
        ],
        style=eval_shared.style_results,
        # className='six columns',
        )
    return children





import os.path
import re

repocess_detect = r'(?P<reprocess>R)?'
base_start = r'(?P<base>capture2?_'
test_type = r'(?P<type>.*)'
timestamp = r'_(?P<date>\d{2}-[A-z][a-z]{2}-\d{2,4})_(?P<time>\d{2}-\d{2}-\d{2})'
access_word = r'_(?P<clip>(?P<talker>[MF]\d)_b(?P<batch>\d+)'\
              r'_w(?P<word>\d)_(?P<cname>[a-z]+))'
suffix = r'(?:_(?P<suffix>BAD|TEMP))?'

match_res = {
                "universal" : repocess_detect +
                              base_start + test_type + timestamp + ')' +
                              '(?:' + access_word + ')?' + suffix,
                "access_csv" : repocess_detect +
                               base_start + test_type + timestamp + ')' +
                               access_word + suffix,
            }

def _normalize(name):
    #check for empty
    if not name:
        raise RuntimeError(f'Unable to get base name from empty string')
    #strip off any path components
    dirname, basename = os.path.split(name)
    #check
    if not basename:
        #name must have ended with a separator
        #get the last path component
        return os.path.basename(dirname)
    else:
        return basename

def match_name(name, re_type="universal", raise_error=True):
    """
    Return a match object for a test filename.

    Match name with one of the regular expressions and return the match object.

    Parameters
    ----------
    name : str
        The name of the measurement file. This can be a fully qualified path or
        just the name of the file.

    re_type : str, default="universal"
        The type of re to use for the match. Currently the only valid values are
        "universal" (default) and "access_csv". Universal should match any name.
        While access_csv will only match access time csv files that contain a
        clip name.

    Returns
    -------
    match object
        The match object for the filename.

    Raises
    ------
    RuntimeError
        If the file name could not be matched.

    KeyError
        If an invalid `re_type` is passed.

    """
    name = _normalize(name)

    m = re.match(match_res[re_type], name)

    if not m and raise_error:
        raise RuntimeError(f'Unable to get base name from \'{name}\'')

    #return match object
    return m

def get_meas_basename(name, re_type="universal"):
    """
    Get the base name for the measurement.

    This returns the base name for a given file name, stripping off any reprocess
    prefix, access time name, or suffix.

    Parameters
    ----------
    name : str
        The name of the measurement file. This can be a fully qualified path or
        just the name of the file.

    re_type : str, default="universal"
        The type of re to use for the match. Currently the only valid values are
        "universal" (default) and "access_csv". Universal should match any name.
        While access_csv will only match access time csv files that contain a
        clip name.

    Returns
    -------
    str
        The base name of the file.

    Raises
    ------
    RuntimeError
        If the file name could not be matched.

    KeyError
        If an invalid `re_type` is passed.
    """

    m = match_name(name, re_type=re_type)

    return m.group('base')

def get_access_clip_info(name):
    """
    Get access time clip information from filename

    Return the talker, batch word and clip name given an access time csv name.

    Parameters
    ----------
    name : str
        The name of access time csv file. This can be a fully qualified path or
        just the name of the file.

    Returns
    -------
    talker : str
        The talker name from the file.
    batch : int
        The batch number for the file.
    word : int
        The word number for the file.
    cname : str
        The clip name for the file.

    Raises
    ------
    RuntimeError
        If the filename could not be matched.
    """

    m = match_name(name, re_type="access_csv")

    return m.group("talker"), int(m.group("batch")), int(m.group("word")), m.group("cname")

def get_meas_date_str(name):
    """"
    Extract date/time from measurement file name.

    Parameters
    ----------
    name : str
        The file name of the measurement. This can be a fully qualified path or
        just the name of the file.

    Returns
    -------
    str
        The date portion of the filename.

    Raises
    ------
    RuntimeError
        If the file name could not be matched.
    """

    m = match_name(name)

    return "_".join(m.group('date','time'))


import os
import fnmatch
import mcvqoe.base
import argparse
import pandas as pd
import math
import re
import numpy as np
import statistics

class Diagnose():
    """
   Diagnose Class to perform diagnostic evaluation of received
    audio files and confirm data integrity.
     
    Use a-weight, FSF, and clipping checks to inform
    user of potential problems in collected data. Flags
    trials that may require further investigation.
    
    Parameters
    ----------
    wav_dir : string
        directory of WAV files
    
    Attributes
    ----------
    tx_filename : list
        names of tx audio files
    tx_wavs   :  list
        audio for tx audio
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings   
    fs : int 
        sampling rate of rx recordings       
    rx_dat : list
        names of rx recordings
    
    Methods
   ----------

    See Also
    --------

    Examples
    --------
    
    Returns
    -------
   
    """
    def __init__(self, 
                 wav_dir = ''):
        self.wav_dir = wav_dir
        # Read in a directory of test trial wav files.
        # Get all the Rx wav files 
        dir_files = os.listdir(self.wav_dir)
        naming = 'Rx*'
        all_wavs = fnmatch.filter(dir_files, naming)
        # Get total number of trials
        self.trials = len(all_wavs)
        # Create empty list for rx recordings 
        self.rx_rec = []
        self.rx_dat = []
        # Cycle through, in order
        # TODO: be clear about testing file formats. This gets confused if
        # the RX numbers reset for each file like in access time
        for n in range(1,self.trials+1):
            start = 'Rx'+str(n)+'_'
            rx_name = [s for s in all_wavs if start in s]
            rx_path = self.wav_dir + '/' + rx_name[0]
            # Check how many channels we have 
            self.fs,y_rec = mcvqoe.base.audio_read(rx_path)
            self.rx_rec.append(y_rec[:]) 
            self.rx_dat.append(rx_name[0])
            # Find all the Tx files in the wav_dir, strip 
            # the Tx and .wav off 
            tx_names = 'Tx*'
            tx_obj = fnmatch.filter(dir_files, tx_names)
            self.tx_filename =fnmatch.filter(tx_obj, '*.wav')
            # Create empty list for tx wavs 
            self.tx_wavs = []
            if tx_obj:
                # Cycle through and get all the TX files
                for k in range(0,len(self.tx_filename)):
                    tx_path = self.wav_dir + '/' + self.tx_filename[k]
                    self.fs,tx_wavfile = mcvqoe.base.audio_read(tx_path)
                    self.tx_wavs.append(tx_wavfile)
            # TODO be robust to scenarios where TX audio is not
            # saved to the data folder        
            # if there are no TX files in the dat dir, find them
    
    def aw_calc(self):
        """
        Calculates the a-weight (dBA) of each trial.
    
        Returns
        -------
        A_Weight : array
            A_Weight of every trial 
        """
        # Create empty list for a-weight values     
        a_weight = []
          
        # Get A-weight
        for k in range(0,self.trials): 
            # Calculate the A-weighted power of each recording 
            aw = mcvqoe.base.a_weighted_power(self.rx_rec[k], self.fs) 
            a_weight.append(aw)
    
        return a_weight
    
    def fsf_calc(self):   
        """
        Calculate FSF scores of each trial. 
    
        Returns
        -------
        FSF_all : list
           FSF scores of every trial 
        """
        # Create empty list for FSF scores
        fsf_all = []
        
        # Cycle through, match RX names to tx names
        # Get just the names of tx files, remove the Tx and .wav 
        tx_base = []
        for n in range(0,len(self.tx_filename)):
            tx_justname = re.sub('\.wav$', '', self.tx_filename[n])
            tx_justname = tx_justname[3:]
            tx_base.append(tx_justname)   
    
        for j in range(0,self.trials): 
            # Find RX files with the matching tx name, create groups 
            match_wavs = re.match(r'(Rx\d+_(?P<tx_base_name>[^.]+))',self.rx_dat[j])
            # find the index of the TX and RX clips to match with the lists of 
            # wav data
            tx_idx = tx_base.index(match_wavs.group('tx_base_name'))
            tx_wav = self.tx_wavs[tx_idx]
            rx_wav = self.rx_rec[j]
            # Get FSF scores for each tx-rx pair
            get_fsf = mcvqoe.base.fsf(rx_wav,tx_wav,self.fs)
            # Get just the FSF score
            fsf_all.append(get_fsf[0])   
    
        return fsf_all

    def peak_amp_calc(self):
        """
        Cycle through all rx recordings and get peak amplitude.

        Returns
        ------- 
        peak_dbfs : list
            peak amplitude of each trial, dB relative to full 
            scale 
        """
        # Create empty list for peak volume
        peak_dbfs = []
        for n in range(0,self.trials):
            # check for positive and negative clipping
            peak = max(abs(self.rx_rec[n]))
            peak_db = round(20 * math.log10(peak), 2)
            peak_dbfs.append(peak_db)
        
        return peak_dbfs
    
    def clip_flag(self,peak_dbfs):
        """
        Cycle through the peak amplitude and check
        for clipping.  
    
        Returns
        -------
        clip_flag : list
            Trials that clipped 
    
        """
        # Set up warning threshold
        vol_high = -1
        # Create empty list for setting flags
        clip_flag = []
        # Check for positive and negative clipping
        peak = np.array(peak_dbfs)
        for n in range(0,self.trials):
            # If approaching clipping, flag as a bad trial
            if peak[n] > vol_high:
                # Flag as a clipped trial
                clip_wav = 1
                # Add it to the bad list
                clip_flag.append(clip_wav) 
            elif peak[n] < vol_high:   
                # Flag as a non-clipped trial
                clip_wav = 0
                # Add it to the bad list
                clip_flag.append(clip_wav)         
        
        return clip_flag   

    def fsf_flag(self,fsf_all):
        """
        Check for trials with an FSF a certain distance
        from the mean. Use this info to find trials that 
        may have lost audio.
            
        Returns
        -------
        fsf_flag : list
            Trials that have low FSF scores or otherwise deviate 
            from the patterns of the dataset
        """
        # Create empty list for FSF flag    
        fsf_flag = []
        # Gather metrics for FSF scores
        fsf_array = np.array(fsf_all)
        fsf_mean = round(statistics.mean(fsf_array),3)
        fsf_std = round(statistics.stdev(fsf_array),3)
        for m in range(0,self.trials):
            # Cycle through FSF scores, look for trials where the score
            # stands out by being a certain distance from the mean
            if abs(fsf_array[m]-fsf_mean) > 2*fsf_std:
                # Flag the trial due to FSF score
                fsf_flag_wav = 1
                # Add it to the bad list
                fsf_flag.append(fsf_flag_wav)
            elif abs(fsf_array[m]-fsf_mean) < 2*fsf_std:
                # Do not flag
                fsf_flag_wav = 0
                # Add it to the bad list
                fsf_flag.append(fsf_flag_wav)   
            # Add a flag for low FSF scores that may be a sign of 
            # dropped audio
            if fsf_array[m] < 0.25:
                # Flag the trial due to low FSF
                fsf_flag_wav = 1
                # Add it to the bad list
                fsf_flag.add(fsf_flag_wav)      
        
        return fsf_flag

    def aw_flag(self,a_weight):
        """
        Check for trials with a dBA less than -60 dBA. Check 
        for trials with a dBA a certain distance from the 
        mean.   
    
        Returns
        -------     
        aw_flag : list
            Trials that have low dBA values (and likely lost audio)
            or otherwise deviate from the patterns of the dataset 
    
        """
        # Create empty list for a-weight flag     
        aw_flag = []
        # Calculate mean a-weight, standard deviation.
        # Use this info to find trials that may have lost
        # audio. 
        aw_array = np.array(a_weight)
        aw_lin = 10**(aw_array/20)
        aw_low = 10**(-60/20)
        aw_mean = round(statistics.mean(aw_lin),3)
        aw_std = round(statistics.stdev(aw_lin),3)
        for m in range(0,self.trials):
            # Cycle through AW values, look for trials where the values
            # stand out by being a certain distance from the mean
            if abs(aw_lin[m]-aw_mean) > 2*aw_std:
                aw_flagged = 1
                # Add it to the bad list
                aw_flag.append(aw_flagged)
            elif abs(aw_lin[m]-aw_mean) < 2*aw_std:
                aw_flagged = 0
                aw_flag.append(aw_flagged)     
            # Add a flag if the a-weight is below -60 dBA
            if aw_lin[m] < aw_low:
                aw_low = 1
                # Add it to the bad list
                aw_flag.add(aw_low)   
        
        return aw_flag
        
    def gather_diagnostics(self,a_weight,fsf_all,peak_dbfs,clip_flag,fsf_flag,aw_flag):
        """
        Create a dataframe of all diagnostic data. A-weight,
        FSF scores, max clip amplitude. Convert to json, csv
        
        Parameters
        ----------
        A_Weight : list    
            A-Weight across all RX trials 
        FSF_all : list
            FSF scores across all RX trials
        peak_dbfs : list     
            Peak amplitude across all RX trials   
        clip_flag : list
            Trials that clipped    
        fsf_flag : list
            Trials that have low FSF scores or otherwise deviate 
            from the patterns of the dataset
        aw_flag : list
            Trials that have low dBA values (and likely lost audio)
            or otherwise deviate from the patterns of the dataset         
        
        Returns
        -------
        diagnostics_csv : csv 
            CSV containing all the dat for diagnostics measurements
       """
       # Create dataframe of info
        df_diagnostics = pd.DataFrame({"RX_Name":self.rx_dat, 
                        "A_Weight":a_weight,
                        "FSF_Scores":fsf_all,
                        "Peak_Amplitude":peak_dbfs,
                        "AW_flag":aw_flag,
                        "Clip_flag":clip_flag,
                        "FSF_flag":fsf_flag})
        # Get session name and use that to name files 
        test_path, test_name =re.split("wav/+", self.wav_dir)

        # Create csv    
        diagnostics_csv = df_diagnostics.to_csv(index=False, line_terminator='\n')
        with open(os.path.join(self.wav_dir,'diagnostics.csv'),'w') as f:
            f.write(diagnostics_csv)
            
        return diagnostics_csv 

def main():   
    """
    Read in a directory of WAV files. 
    Use a-weight, FSF, and clipping checks to inform
    user of potential problems in collected data. Flags
    trials that may require further investigation.

    Parameters
    ----------
    Wav_Dir : string
        directory of WAV files

    Returns
    -------

    """
    # Input parsing
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "wav_dir",
        default=None,
        type=str,
        help="Directory to test data wav files",
    )
    # Parse input arguments
    args = parser.parse_args()
    obj = Diagnose(args.wav_dir)
    print("Measuring a-weight") 
    a_weight = obj.aw_calc()   
    aw_flag = obj.aw_flag(a_weight)
    print("Measuring FSF") 
    fsf_all = obj.fsf_calc()
    fsf_flag = obj.fsf_flag(fsf_all)
    print("Measuring peak amplitude") 
    peak_dbfs = obj.peak_amp_calc()
    clip_flag = obj.clip_flag(peak_dbfs)
    print("Creating json and csv") 
    obj.gather_diagnostics(a_weight,fsf_all,peak_dbfs,clip_flag,fsf_flag,aw_flag)

if __name__ == "__main__":
    main()



from .IRIGB_decode import IRIGB_decode
from .soft_timecode import soft_time_decode

'''
Types of timecodes and their decode function
'''
timecode_types = {'IRIGB_timecode' : IRIGB_decode,'soft_timecode' : soft_time_decode}

def time_decode(tc_type, timecode_audio, fs, decode_args={}):
    '''
    Decode a timecode of the given type.
    '''
    try:
        tc_func=timecode_types[tc_type]
    except KeyError:
        raise ValueError(f'\'{tc_type}\' must be one of {list(timecode_types.keys())}')
    
    return tc_func(timecode_audio,fs,**decode_args)


import numpy as np
from scipy.cluster.vq import kmeans2
from warnings import warn
import datetime


  

# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST
# employees are not subject to copyright protection in the United States and
# are considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty
# appears in all copies.

# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY
# WARRANTY THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# FREEDOM FROM INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL
# CONFORM TO THE SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR
# FREE. IN NO EVENT SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
# LIMITED TO, DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING
# OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THIS SOFTWARE,
# WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER
# OR NOT INJURY WAS SUSTAINED BY PERSONS OR PROPERTY OR OTHERWISE, AND
# WHETHER OR NOT LOSS WAS SUSTAINED FROM, OR AROSE OUT OF THE RESULTS OF, OR
# USE OF, THE SOFTWARE OR SERVICES PROVIDED HEREUNDER.

def IRIGB_decode(tca, fs, tc_tol=0.05):
    '''
    decodes an IRIG-B timecode
    
    Parameters
    ----------
    tca : numpy array
        timecode audio data
    fs : int
        the audio sample rate for tca.
    tc_tol: float, default=0.05
        Time code tolerance. This changes the thresholds for what bit periods
        are considered a one, zero and frame marker.
    
    Returns
    -------
    dates : tuple of datetime
        Tuple of decode times.
    fsamp : list of ints
        Sample numbers that decoded times came from.
    '''
    
    #make sure that time code input is a numpy array
    tca = np.array(tca, dtype="float64")

    #tc_tol must be in [0, 0.5]
    if not (isinstance(tc_tol, (int, float)) and tc_tol >= 0 and tc_tol <= 0.5):
        raise ValueError(f'tc_tol must be > 0 and < 0.5. Got tc_tol = {tc_tol}')

    #calculate the upper envelope
    env = envelope(tca, 40)

    #use kmeans to threshold the envelope into 2 clusters
    #use 0 and max as the starting centroids
    init = (0, 1)
    #take the cluster indices, raise an error if we don't have 2 centroids
    _, env_th = kmeans2(env, init, missing='raise')

    #find edges assume that signal starts high so that we see the first real
    edges = np.append([0], 1+np.nonzero(np.diff(env_th))[0])
    #edges = 1 + np.where(edges!=0)[0]

    #rising edges
    r_edg = edges[env_th[edges] == 1] - 1

    #falling edges
    f_edg = np.append([1], edges[env_th[edges] == 0][1:]) - 1

    #get the first and last rising edge
    startRise = r_edg[0]
    endRise = r_edg[-1]

    #remove falling edges that happen before the first rising edge and after
    #the last
    f_edg = f_edg[np.logical_and(f_edg > startRise, f_edg < endRise)]

    #calculate period
    T = np.diff(r_edg) / fs

    #calculate the pulse width
    #there is one more rising edge than falling edge so ignore last rising edge
    pw = (f_edg - r_edg[:-1]) / fs

    bits = pw_to_bits(pw, 10e-3, tc_tol)
    Tbit = 10e-3

    #find invalid periods
    invalid = np.logical_or((T < Tbit * (1-tc_tol)), (T > Tbit *(1+tc_tol)))


    #mark bits with invalid periods as invalid
    for x in range(0, len(invalid)):
        if invalid[x]:
            bits[x] = -2
    #index within a frame -1 means invalid frame
    frame_idx=-1

    weight=[  1,  2,  4,  8,  0, 10, 20, 40, -1,  1,  2,  4,  8,  0, 10, 20, 40,  0, -1,
          1,  2,  4,  8,  0, 10, 20,  0,  0, -1,  1,  2,  4,  8,  0, 10, 20, 40, 80, -1,
        100,200,  0,  0,  0,0.1,0.2,0.4,0.8, -1,  1,  2,  4,  8,  0, 10, 20, 40, 80, -1,
          1,  2,  4,  8, 16, 32, 64,128,256, -1,  1,  2,  4,  8, 16, 32, 64,128,256, -1,
          1,  2,  4,  8, 16, 32, 64,128,256, -1,512,1024,2048,4096,8192,16384,32768,65536,0,-1]

    value=[  1,1,1,1,1,1,1,1,-1,2,2,2,2,2,2,2,2,2,-1,
           3,3,3,3,3,3,3,3,3,-1,4,4,4,4,4,4,4,4,4,-1,
           4,4,4,4,4,5,5,5,5,-1,6,6,6,6,6,6,6,6,6,-1,
           7,7,7,7,7,7,7,7,7,-1,8,8,8,8,8,8,8,8,8,-1,
           9,9,9,9,9,9,9,9,9,-1,9,9,9,9,9,9,9,9,9,-1]

    #preallocate frames to hold the maximum number of frames
    frames = np.zeros((int(np.floor(len(bits)/100)), max(value)))
    #preallocate frame bits
    fbits = np.zeros((int(np.floor(len(bits)/100)), 100))
    #sample number of the first rising edge in the frame
    fsamp = np.zeros((int(np.floor(len(bits)/100)),),dtype=np.int)

    #frame number
    fnum=0

    for k in range(1, len(bits)):
        #check if current flame is invalid
        if(frame_idx==-1):
            #check for a frame start
            if(bits[k]==2 and bits[k-1]==2):
                #set new frame index
                frame_idx=0
                #zero frame data
                frame=np.zeros(max(value))
        else:
            #check if this should be a marker bit
            if(frame_idx % 10) == 8:
                #check if marker is found
                if (bits[k]!=2):
                    #give warning for missing marker
                    warn('Marker not found at frame index %i' % frame_idx)
                    #reset frame index
                    frame_idx=-1
                    #restart loop
                    continue
            else:
                #check that bit is a 1 or zero
                if(bits[k]!=1 and bits[k]!=0):
                    #give warning based on bit value
                    bit = bits[k]
                    if (bit == -2):
                        #give warning for invalid period
                        warn('Invalid bit period at frame index %i' % frame_idx)
                    elif(bit == -1):
                        #give warning for invalid bit value
                        warn('Invalid bit at frame index %i' % frame_idx)
                            
                    elif(bit == 2):
                        #give warning for unexpected marker
                        warn('Unexpected marker at frame index %i' % frame_idx)
                    else:
                        #give warning for invalid bit value
                        warn('Unexpected bit value %i at frame index %i' %(bits[k],frame_idx))
                    #reset frame index
                    frame_idx=-1
                    #restart loop
                    continue
                #get value idx
                vi=value[frame_idx]
                #otherwise get bit value
                frame[vi-1]=frame[vi-1]+bits[k]*weight[frame_idx]
            #increment frame index
            frame_idx=frame_idx+1
            #check if frame is complete
            if(frame_idx>=99):
                #store decoded frame data
                frames[fnum]=frame
                #store decoded frame bits
                fbits[fnum]=bits[(k-99):k+1]
                #get sample number of the first rising edge after frame marker
                fsamp[fnum]=r_edg[k-98]
                #search for next frame
                frame_idx=-1
                #increment frame number
                fnum=fnum+1

    #remove extra data
    frames = frames[0:fnum]
    fbits = fbits[0:fnum]
    fsamp = fsamp[0:fnum]



    #create date vectors
    #IRIG B does not give month or day of month so use a empty part of the
    #frame for month and fix it to 1 later. Day of year will wrap to the
    #correct month

    dvec=frames[:,[5,4,3,2,1,0]]
    
    #add in year digits from current year
    dvec[:,0]= dvec[:,0] + np.floor(datetime.datetime.now().year / 100) * 100
    
    #set month to 1
    dvec[:,1]=1

    dates = []
    for d in dvec:
        year, month, day, hour, minute, second = tuple(int(d[i]) for i in (0, 1, 2, 3, 4, 5))
        date = datetime.datetime(year, month, 1, hour, minute, second)
        #create a dummy time delta of the days, which will be added to each time
        #make sure to subtract 1 from days because default day value of 1 was used
        dummyTime = datetime.timedelta(days=day-1)
        #format the datetime into a string
        dates += [date + dummyTime]

    #this version returns the dates as a string instead of a datetime object
    #this can be used for testing purposes
    '''dates = []
    for d in dvec:
        year, month, day, hour, minute, second = tuple(int(d[i]) for i in (0, 1, 2, 3, 4, 5))
        date = datetime.datetime(year, month, 1, hour, minute, second)
        #create a dummy time delta of the days, which will be added to each time
        #make sure to subtract 1 from days because default day value of 1 was used
        dummyTime = datetime.timedelta(days=day-1)
        #format the datetime into a string
        dates += [(date + dummyTime).strftime("%d-%b-%Y %H:%M:%S")]'''

    return np.array(dates), fsamp


    
    
#helper function used in envelope()
#calulates amplitudes in singal x using N-tap Hilbert filter
def envFIR(x, n):

    #construct ideal hilbert filter truncated to desired length
    fc = 1
    t = (fc/2) * np.arange((1-n)/2, ((n-1)/2) + 1)
    
    hfilt = np.sinc(t) * np.exp(1j*np.pi * t)

    #multiply ideal filter with tapered window
    beta = 8
    firFilter = hfilt * np.kaiser(n,beta)
    firFilter = firFilter / sum(np.real(firFilter))

    #apply filter and take the magnitude
    y = np.abs(np.convolve(x,firFilter,'same'))
    return(y)
    

    #although the following code affected the output of envFIR, it did not affect
    #the final result of time_decode as it is negligible
    '''Due to differences in convolution algorithims for even-length inputs
    in python and MATLAB, we must delete the first sample data point of the 
    signal, and append an additonal sample data point onto the end - 
    calculated from the result of the 'full' convolution. 
    This yields the MATLAB equivalent
    y = y[1:]
    temp = np.abs(np.convolve(x,firFilter,'full')).tolist()
    y = np.append(y, temp[temp.index(y[-1])+1])
    return(y)'''


#uses an N-tap Hilbert filter to compute the upper envelope of X
def envelope(x, n):
    xmean = np.mean(x)
    xcentered = x - xmean
    xampl = envFIR(xcentered, n)
    yupper = xmean + xampl

    return yupper


#validates the pulse width with an array of bools
def is_valid_pw(val,th):
    return np.logical_and(val > th[0], val < th[1])

#helps to ensure that thresholds don't overlap
def fix_overlap(t1,t2):
    #check if thresholds overlap    
    if(t1>t2):
        #set thresholds to average
        t1=np.mean([t1,t2])
        #this calculates machine epsilon
        t2=t1+np.finfo(float).eps
    return t1, t2


def pw_to_bits(pw,Tb,tol):
    #thresholds for ones
    Th1 = 0.5 * Tb + Tb * np.array([-tol,tol])
    #thresholds for zeros
    Th0 = 0.2 * Tb + Tb * np.array([-tol,tol])
    #thresholds for marker
    Thm = 0.8 * Tb + Tb * np.array([-tol,tol])
    #make sure thresholds don't overlap
    Th0[1],Th1[0] = fix_overlap(Th0[1],Th1[0])
    Th1[1],Thm[0] = fix_overlap(Th1[1],Thm[0])
    #check for valid pulse width for a one
    valid1 = is_valid_pw(pw,Th1)
    #check for valid pulse width for a zero
    valid0 = is_valid_pw(pw,Th0)
    #check for valid marker pulse width
    validmk = is_valid_pw(pw,Thm)
    #print(validmk.tolist())
    
    #return 0 for zero 1 for one 2 for mark and -1 for invalid
    bits = valid1 + (2 * validmk) - np.logical_not(np.logical_or.reduce((validmk, valid1, valid0)))
    return bits


import csv
import glob
import datetime
import os
import json
import mcvqoe.timing
import re
import shutil
import string
import time
import zipfile

import numpy as np

from itertools import cycle
from .misc import audio_write, write_cp
from .naming import get_meas_basename
from .write_log import fill_log, pre as log_pre, post as log_post

class Measure:

    no_log = ()

    #measurement name, overide in subclass
    measurement_name = "Base"

    required_chans = {
                    '1loc' : {
                                "rec" : ("rx_voice",),
                                "pb" : ("tx_voice",),
                             },
                    # NOTE : for 2 location, recording inputs will be checked
                    #        to see if they include a timecode channel
                    '2loc_tx' : {
                                "rec" : (),
                                "pb" : ("tx_voice",),
                                },
                    '2loc_rx' : {
                                "rec" : ("rx_voice",),
                                "pb" : (),
                             },
                      }

    #filename for zipped audio
    _zip_name = 'audio.zip'

    @staticmethod
    def channel_check(expected, given):
        missing = []
        for name in expected:
            if name not in given:
                missing.append(name)
        return missing

    def check_channels(self):
        rec_missing = self.channel_check(
                                    self.required_chans[self.test]['rec'],
                                    self.audio_interface.rec_chans.keys()
                                         )
        if rec_missing:
            raise ValueError(f"self.audio_interface missing recording channels for : {rec_missing}")

        pb_missing = self.channel_check(
                                    self.required_chans[self.test]['pb'],
                                    self.audio_interface.playback_chans.keys()
                                        )
        if rec_missing:
            raise ValueError(f"self.audio_interface missing playback channels for : {pb_missing}")

    def audio_clip_check(self):
        #dummy function, override if needed
        pass

    def log_extra(self):
        """
        A place to add test specific fields to the log
        """
        #dummy function, override if needed
        pass

    def param_check(self):
        """
        Check that parameters are correct.
        
        Raises
        ------
        ValueError
            If there is an incorrect parameter.
        """
        #dummy function, override if needed
        pass

    def test_setup(self):
        """
        Extra things that need to be setup for a specific test
        """
        #dummy function, override if needed
        pass

    def run(self, **kwargs):
        if self.test == "1loc":
            return self.run_1loc(**kwargs)
        elif self.test == "2loc_tx":
            return self.run_2loc_tx(**kwargs)
        elif self.test == "2loc_rx":
            return self.run_2loc_rx(**kwargs)
        else:
            raise ValueError(f'Unknown test type "{self.test}"')

    def run_1loc(self):
        """
        Run a generic test.
        """
        # ------------------------[Test specific setup]------------------------
        self.test_setup()
        # ------------------[Check for correct audio channels]------------------
        self.check_channels()
        # -------------------------[Get Test Start Time]-------------------------

        self.info["Tstart"] = datetime.datetime.now()
        dtn = self.info["Tstart"].strftime("%d-%b-%Y_%H-%M-%S")

        # --------------------------[Fill log entries]--------------------------
        # set test name
        self.info["test"] = self.measurement_name
        #add any extra entries
        self.log_extra()
        # fill in standard stuff
        self.info.update(fill_log(self))

        # -----------------------[Setup Files and folders]-----------------------

        # generate data dir names
        data_dir = os.path.join(self.outdir, "data")
        wav_data_dir = os.path.join(data_dir, "wav")
        csv_data_dir = os.path.join(data_dir, "csv")

        # create data directories
        os.makedirs(csv_data_dir, exist_ok=True)
        os.makedirs(wav_data_dir, exist_ok=True)

        # generate base file name to use for all files
        base_filename = "capture_%s_%s" % (self.info["Test Type"], dtn)

        # generate test dir names
        wavdir = os.path.join(wav_data_dir, base_filename)

        # create test dir
        os.makedirs(wavdir, exist_ok=True)

        # generate csv name
        self.data_filename = os.path.join(csv_data_dir, f"{base_filename}.csv")

        # generate temp csv name
        temp_data_filename = os.path.join(csv_data_dir, f"{base_filename}_TEMP.csv")

        # ---------------------[Load Audio Files if Needed]---------------------

        if not hasattr(self, "y"):
            self.load_audio()

        #check audio clips, and possibly, adjust the number of trials
        self.audio_clip_check()

        # generate clip index
        self.clipi = self.rng.permutation(self.trials) % len(self.y)

        # -----------------------[Add Tx audio to wav dir]-----------------------

        # get name with out path or ext
        clip_names = [os.path.basename(os.path.splitext(a)[0]) for a in self.audio_files]


        if hasattr(self, 'cutpoints'):
            cutpoints = self.cutpoints
        else:
            #placeholder for zip
            cutpoints = cycle((None,))
        # write out Tx clips to files
        # cutpoints, if present, are always written
        for dat, name, cp in zip(self.y, clip_names, cutpoints):
            out_name = os.path.join(wavdir, f"Tx_{name}")
            if self.save_tx_audio and self.save_audio:
                audio_write(out_name + ".wav", int(self.audio_interface.sample_rate), dat)
            #write cutpoints, if present
            if cp:
                write_cp(out_name+'.csv',cp)


        # -------------------------[Generate CSV header]-------------------------

        header, dat_format = self.csv_header_fmt()

        # ---------------------------[write log entry]---------------------------

        log_pre(info=self.info, outdir=self.outdir)

        # ---------------[Try block so we write notes at the end]---------------

        try:

            # -------------------------[Turn on RI LED]-------------------------
            self.ri.led(1, True)

            # -----------------------[write initial csv file]-----------------------
            with open(temp_data_filename, "wt") as f:
                f.write(header)

            # ------------------------[Measurement Loop]------------------------

            #zero pause count
            self._pause_count = 0

            if not hasattr(self, 'pause_trials'):
                # if we don't have pause_trials, that means no pauses
                self.pause_trials = np.inf

            for trial in range(self.trials):
                # -----------------------[Update progress]-------------------------
                if not self.progress_update("test", self.trials, trial):
                    # turn off LED
                    self.ri.led(1, False)
                    print("Exit from user")
                    break
                # -----------------------[Get Trial Timestamp]-----------------------
                ts = datetime.datetime.now().strftime("%d-%b-%Y %H:%M:%S")

                # --------------------[Key Radio and play audio]--------------------

                # Press the push to talk button
                self.ri.ptt(True)

                # Pause the indicated amount to allow the radio to access the system
                time.sleep(self.ptt_wait)

                clip_index = self.clipi[trial]

                # Create audiofile name/path for recording
                audioname = f"Rx{trial+1}_{clip_names[clip_index]}.wav"
                audioname = os.path.join(wavdir, audioname)

                # Play/Record
                rec_chans = self.audio_interface.play_record(self.y[clip_index], audioname)

                # Release the push to talk button
                self.ri.ptt(False)

                # -----------------------[Pause Between runs]-----------------------

                time.sleep(self.ptt_gap)

                # -----------------------------[Data Processing]----------------------------

                trial_dat = self.process_audio(
                    clip_index,
                    audioname,
                    rec_chans,
                )

                # add extra info
                trial_dat["Timestamp"] = ts
                trial_dat["Filename"] = clip_names[clip_index]
                trial_dat['Over_runs']  = 0
                trial_dat['Under_runs'] = 0

                # -------------------[Delete file if needed]-------------------
                if not self.save_audio:
                    os.remove(audioname)

                # --------------------------[Write CSV]--------------------------

                with open(temp_data_filename, "at") as f:
                    f.write(dat_format.format(**trial_dat))


                #------------------[Check if we should pause]------------------

                #increment pause count
                self._pause_count += 1

                if self._pause_count >= self.pause_trials:

                    #zero pause count
                    self._pause_count = 0

                    # Calculate set time
                    time_diff = datetime.datetime.now().replace(microsecond=0)
                    set_time = time_diff - set_start

                    # Turn on LED when waiting for user input
                    self.ri.led(2, True)

                    # wait for user
                    user_exit = self.user_check(
                            'normal-stop',
                            'check batteries.',
                            trials=self.pause_trials,
                            time=set_time,
                        )

                    # Turn off LED, resuming
                    self.ri.led(2, False)

                    if(user_exit):
                        raise SystemExit()

                    # Save time for next set
                    set_start = datetime.datetime.now().replace(microsecond=0)

            # -----------------------------[Cleanup]-----------------------------

            # move temp file to real file
            shutil.move(temp_data_filename, self.data_filename)

            # ---------------------------[Turn off RI LED]---------------------------

            self.ri.led(1, False)

        finally:
            if self.get_post_notes:
                # get notes
                info = self.get_post_notes()
            else:
                info = {}
            # finish log entry
            log_post(outdir=self.outdir, info=info)

        #return filename in a list
        return (self.data_filename,)

    def run_2loc_tx(self):
        """
        Run a two location test.

        This is a generic test function for tests like m2e, PSuD and
        Intelligibility.
        """

        # ------------------------[Test specific setup]------------------------
        self.test_setup()
        # ------------------[Check for correct audio channels]------------------
        self.check_channels()
        # we need to be recording a timecode
        mcvqoe.timing.require_timecode(self.audio_interface)
        # -------------------------[Get Test Start Time]-------------------------

        self.info["Tstart"] = datetime.datetime.now()
        dtn = self.info["Tstart"].strftime("%d-%b-%Y_%H-%M-%S")

        # --------------------------[Fill log entries]--------------------------
        # set test name, needs to match log_search.datafilenames
        self.info["test"] = "Tx Two Loc Test"
        #add any extra entries
        self.log_extra()
        # fill in standard stuff
        self.info.update(fill_log(self))

        # -----------------------[Setup Files and folders]-----------------------

        # generate data dir names
        data_dir = os.path.join(self.outdir, "data")
        tx_dat_fold = os.path.join(data_dir, "2loc_tx-data")

        # generate base file name to use for all files
        base_filename = "capture_%s_%s" % (self.info["Test Type"], dtn)

        wavdir = os.path.join(tx_dat_fold, "Tx_" + base_filename)

        # create directories
        os.makedirs(wavdir, exist_ok=True)

        # Put .csv files in wav dir
        csv_data_dir = wavdir

        # generate csv name
        self.data_filename = os.path.join(csv_data_dir, f"{base_filename}.csv")

        # generate temp csv name
        temp_data_filename = os.path.join(csv_data_dir, f"{base_filename}_TEMP.csv")

        # ---------------------[Load Audio Files if Needed]---------------------

        if not hasattr(self, "y"):
            self.load_audio()

        self.audio_clip_check()

        # generate clip index
        self.clipi = self.rng.permutation(self.trials) % len(self.y)

        # -----------------------[Add Tx audio to wav dir]-----------------------

        # get name with out path or ext
        clip_names = [os.path.basename(os.path.splitext(a)[0]) for a in self.audio_files]

        if hasattr(self, 'cutpoints'):
            cutpoints = self.cutpoints
        else:
            #placeholder for zip
            cutpoints = cycle((None,))

        # write out Tx clips to files
        for dat, name, cp in zip(self.y, clip_names, cutpoints):
            out_name = os.path.join(wavdir, f"Tx_{name}")
            audio_write(out_name + ".wav", int(self.audio_interface.sample_rate), dat)
            #write cutpoints, if present
            if cp:
                write_cp(out_name+'.csv',cp)

        # -------------------------[Generate CSV header]-------------------------

        header, dat_format = self.csv_header_fmt()

        # ---------------------------[write log entry]---------------------------

        log_pre(info=self.info, outdir=self.outdir)

        # ---------------[Try block so we write notes at the end]---------------
        try:

            # -----------------------[write initial csv file]-----------------------
            with open(temp_data_filename, "wt") as f:
                f.write(header)

            # -------------------------[Turn on RI LED]-------------------------

            self.ri.led(1, True)

            # ------------------------[Measurement Loop]------------------------

            #zero pause count
            self._pause_count = 0

            if not hasattr(self, 'pause_trials'):
                # if we don't have pause_trials, that means no pauses
                self.pause_trials = np.inf

            for trial in range(self.trials):

                # -----------------------[Update progress]-------------------------
                if not self.progress_update("test", self.trials, trial):
                    # turn off LED
                    self.ri.led(1, False)
                    print("Exit from user")
                    break
                # -----------------------[Get Trial Timestamp]-----------------------
                ts = datetime.datetime.now().strftime("%d-%b-%Y %H:%M:%S")

                # --------------------[Key Radio and play audio]--------------------

                # Press the push to talk button
                self.ri.ptt(True)

                # Pause the indicated amount to allow the radio to access the system
                time.sleep(self.ptt_wait)

                clip_index = self.clipi[trial]

                # Create audiofile name/path for recording
                audioname = f"Rx{trial+1}_{clip_names[clip_index]}.wav"
                audioname = os.path.join(wavdir, audioname)

                # Play/Record
                rec_chans = self.audio_interface.play_record(self.y[clip_index], audioname)

                # Release the push to talk button
                self.ri.ptt(False)

                # -----------------------[Pause Between runs]-----------------------

                time.sleep(self.ptt_gap)

                # --------------------------[Write CSV]--------------------------

                chan_str = "(" + (";".join(rec_chans)) + ")"

                #generate dummy values for format
                trial_dat = {}
                for _, field, _, _ in string.Formatter().parse(dat_format):
                    if field not in self.data_fields:
                        if field is None:
                            #we got None, skip this one
                            continue
                        #check for array
                        m = re.match(r'(?P<name>.+)\[(?P<index>\d+)\]',field)
                        if not m:
                            #not in data fields, fill with NaN
                            trial_dat[field] = np.NaN
                        else:
                            field_name = m.group("name")
                            index = int(m.group("index"))
                            if field_name not in trial_dat or \
                                len(trial_dat[field_name]) < index + 1:
                                trial_dat[field_name] = (np.NaN,) * (index +1)
                    elif self.data_fields[field] is float:
                        #float, fill with NaN
                        trial_dat[field] = np.NaN
                    elif self.data_fields[field] is int:
                        #int, fill with zero
                        trial_dat[field] = 0
                    else:
                        #something else, fill with None
                        trial_dat[field] = None

                #fill in known values
                trial_dat['Timestamp'] = ts
                trial_dat['Filename'] = clip_names[clip_index]
                trial_dat['channels'] = chan_str

                with open(temp_data_filename, "at") as f:
                    f.write(
                        dat_format.format(
                            **trial_dat
                        )
                    )

                #------------------[Check if we should pause]------------------

                #increment pause count
                self._pause_count += 1

                if self._pause_count >= self.pause_trials:

                    #zero pause count
                    self._pause_count = 0

                    # Calculate set time
                    time_diff = datetime.datetime.now().replace(microsecond=0)
                    set_time = time_diff - set_start

                    # Turn on LED when waiting for user input
                    self.ri.led(2, True)

                    # wait for user
                    user_exit = self.user_check(
                            'normal-stop',
                            'check batteries.',
                            trials=self.pause_trials,
                            time=set_time,
                        )

                    # Turn off LED, resuming
                    self.ri.led(2, False)

                    if(user_exit):
                        raise SystemExit()

                    # Save time for next set
                    set_start = datetime.datetime.now().replace(microsecond=0)

            # -----------------------------[Cleanup]-----------------------------

            # move temp file to real file
            shutil.move(temp_data_filename, self.data_filename)

            # ---------------------------[Turn off RI LED]---------------------------

            self.ri.led(1, False)

            # -----------------------[Notify User of Completion]------------------------

            self.progress_update(
                "status",
                self.trials,
                self.trials,
                msg="Data collection complete, you may now stop data collection on" + " the receiving end",
            )

        finally:
            if self.get_post_notes:
                # get notes
                info = self.get_post_notes()
            else:
                info = {}
            # finish log entry
            log_post(outdir=self.outdir, info=info)

        #return filename in a list
        return (self.data_filename,)

    def run_2loc_rx(self):
        """
        Two location recive basic test.

        This function just records audio with a little bit of logging. Should be
        usable by all 2 location tests.
        """

        # ------------------------[Test specific setup]------------------------
        self.test_setup()
        # ------------------[Check for correct audio channels]------------------
        self.check_channels()
        # we need to be recording a timecode
        mcvqoe.timing.require_timecode(self.audio_interface)

        # -------------------------[Get Test Start Time]-------------------------
        self.info["Tstart"] = datetime.datetime.now()
        dtn = self.info["Tstart"].strftime("%d-%b-%Y_%H-%M-%S")

        # --------------------------[Fill log entries]--------------------------

        # set test name, needs to match log_search.datafilenames
        self.info["test"] = "Rx Two Loc Test"
        #add any extra entries
        self.log_extra()
        # fill in standard stuff
        self.info.update(fill_log(self))

        # -----------------------[Setup Files and folders]-----------------------

        # Create rx-data folder
        data_dir = os.path.join(self.outdir, "data")
        rx_dat_fold = os.path.join(data_dir, "2loc_rx-data")
        os.makedirs(rx_dat_fold, exist_ok=True)

        base_filename = "capture_%s_%s" % (self.info["Test Type"], dtn)

        self.data_filename = os.path.join(rx_dat_fold, "Rx_" + base_filename + ".wav")

        info_name = os.path.join(rx_dat_fold, "Rx_" + base_filename + ".json")

        # ---------------------------[write log entry]---------------------------

        log_pre(info=self.info, outdir=self.outdir)

        # ---------------[Try block so we write notes at the end]---------------
        try:
            # ----------------------[Send progress update]---------------------
            self.progress_update("status", 1, 0, msg="Two location receive recording running")

            # --------------------------[Record audio]--------------------------
            rec_names = self.audio_interface.record(self.data_filename)

            # ------------------------[Save audio info]------------------------

            with open(info_name, "wt") as info_f:
                json.dump({"channels": rec_names}, info_f)

        finally:
            if self.get_post_notes:
                # get notes
                info = self.get_post_notes()
            else:
                info = {}
            # finish log entry
            log_post(outdir=self.outdir, info=info)

        return (self.data_filename,)

    def load_test_data(self, fname, load_audio=True, audio_path=None):
        """
        load test data from .csv file.

        Parameters
        ----------
        fname : string
            filename to load
        load_audio : bool, default=True
            if True, finds and loads audio clips and cutpoints based on fname
        audio_path : str, default=None
            Path to find audio files at. Guessed from fname if None.

        Returns
        -------
        list of dicts
            returns data from the .csv file

        """

        with open(fname, "rt") as csv_f:
            # create dict reader
            reader = csv.DictReader(csv_f)
            print(f"reader header: {reader.fieldnames}")
            # create empty list
            data = []
            # create set for audio clips
            clips = set()
            for row in reader:
                # convert values proper datatype
                for k in row:
                    # check for clip name
                    if k == "Filename":
                        # save clips
                        clips.add(row[k])
                    try:
                        # check for None field
                        if row[k] == "None":
                            # handle None correctly
                            row[k] = None
                        else:
                            # convert using function from data_fields
                            row[k] = self.data_fields[k](row[k])
                    except KeyError:
                        # not in data_fields, convert to float
                        row[k] = float(row[k])

                # append row to data
                data.append(row)

        # set total number of trials, this gives better progress updates
        self.trials = len(data)

        # check if we should load audio
        if load_audio:
            #we do not want to load the full dir for reprocessing
            self.full_audio_dir = False

            # set audio file names to Tx file names
            self.audio_files = ["Tx_" + name + ".wav" for name in clips]

            print(f'Audio clip names : {self.audio_files}')

            dat_name = get_meas_basename(fname)

            if audio_path is not None:
                self.audio_path = audio_path
            else:
                # set audio_path based on filename
                self.audio_path = os.path.join(os.path.dirname(os.path.dirname(fname)), "wav", dat_name)

            # load audio data from files
            self.load_audio()
            self.audio_clip_check()

        return data

    # get the clip index given a partial clip name
    def find_clip_index(self, name):
        """
        find the inex of the matching transmit clip.

        Parameters
        ----------
        name : string
            base name of audio clip

        Returns
        -------
        int
            index of matching tx clip

        """

        # match a string that has the chars that are in name
        name_re = re.compile(re.escape(name) + "(?![^.])")
        # get all matching indices
        match = [idx for idx, clip in enumerate(self.audio_files) if name_re.search(clip)]
        # check that a match was found
        if not match:
            raise RuntimeError(f"no audio clips found matching '{name}' found in {self.audio_files}")
        # check that only one match was found
        if len(match) != 1:
            raise RuntimeError(f"multiple audio clips found matching '{name}' found in {self.audio_files}")
        # return matching index
        return match[0]

    @staticmethod
    def unzip_audio(audio_path):
        zip_path = os.path.join(audio_path,Measure._zip_name)
        if zipfile.is_zipfile(zip_path):
            audio_zip = zipfile.ZipFile(zip_path,mode='r')
            #extract all files into the audio dir
            audio_zip.extractall(audio_path)

    def zip_wavdir(self, path):
        """
        Replace the receive audio files in `path` with a zip file.

        Parameters
        ----------
        path : string
            A path to the directory where the test .wav files are stored.

        """
        with zipfile.ZipFile(
                    os.path.join(path,self._zip_name),
                    mode='w',
                    compression=zipfile.ZIP_LZMA,
                ) as audio_zip:
            #find all the rx wav files
            rx_wavs = glob.glob(os.path.join(path,'Rx*.wav'))
            #fid all bad files
            bad_wavs = glob.glob(os.path.join(path,'Bad*.wav'))
            #zip bad files and Rx files
            zip_wavs = rx_wavs + bad_wavs
            #get number of files
            num_zip_files = len(zip_wavs)
            for n, name in enumerate(zip_wavs):
                bname =  os.path.basename(name)
                self.progress_update('compress',num_zip_files,n)
                audio_zip.write(name,arcname=bname)

        #zip file has been written, delete files
        self.progress_update('status',num_zip_files,num_zip_files,msg='Deleting compressed audio...')
        for name in zip_wavs:
            os.remove(name)

    def post_process(self, test_dat, fname, audio_path):
        """
        process csv data.

        Parameters
        ----------
        test_data : list of dicts
            csv data for trials to process
        fname : string
            file name to write processed data to
        audio_path : string
            where to look for recorded audio clips

        Returns
        -------

        """

        #do extra setup things
        self.test_setup()

        # get .csv header and data format
        header, dat_format = self.csv_header_fmt()

        with open(fname, "wt") as f_out:

            f_out.write(header)

            for n, trial in enumerate(test_dat):

                # update progress
                self.progress_update("proc", self.trials, n)

                # find clip index
                clip_index = self.find_clip_index(trial["Filename"])
                # create clip file name
                clip_name = "Rx" + str(n + 1) + "_" + trial["Filename"] + ".wav"

                try:
                    # attempt to get channels from data
                    rec_chans = trial["channels"]
                except KeyError:
                    # fall back to only one channel
                    rec_chans = ("rx_voice",)
                new_dat = self.process_audio(
                        clip_index,
                        os.path.join(audio_path, clip_name),
                        rec_chans
                        )

                # overwrite new data with old and merge
                merged_dat = {**trial, **new_dat}

                # write line with new data
                f_out.write(dat_format.format(**merged_dat))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mcvqoe-base",
    author="PSCR",
    author_email="PSCR@PSCR.gov",
    description="Common code for MCV QoE Measurement Classes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/usnistgov/mcvqoe-base",
    packages=setuptools.find_namespace_packages(include=["mcvqoe.*"]),
    include_package_data=True,
    package_data={
        "mcvqoe": ["audio_clips", "test.wav"],
        "mcvqoe": ["gui", "MCV-sm.png"],
    },
    use_scm_version={"write_to": "mcvqoe/base/version.py"},
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    license="NIST software License",
    install_requires=[
        "numpy",
        "scipy",
        "sounddevice",
        "pyserial",
        "soundfile",
        "appdirs",
        'importlib-metadata ; python_version < "3.8"',
        "pandas",
    ],
    entry_points={
        "console_scripts": [
            "testCpy=mcvqoe.utilities.test_copy:main",
            "local-copy=mcvqoe.utilities.local_copy:main",
            "mcvqoe-test-play=mcvqoe.hardware.PTT_play:main",
            "mcvqoe-chan=mcvqoe.utilities.channelinfo:main",
            "mcvqoe-2loc-process=mcvqoe.timing.two_loc_process:main",
            "mcvqoe-reprocess=mcvqoe.utilities.reprocess:main",
            "mcvqoe-diagnostics=mcvqoe.diagnostics.diagnostics:main",
        ],
        "mcvqoe.channel": "clean=mcvqoe.simulation.cleanchan",
        "mcvqoe.impairment": "probabilityiser=mcvqoe.simulation.pbiplugin",
    },
    python_requires=">=3.6",
)
